import math
import logging
import random
import uuid
from abc import ABC, abstractmethod
from easydict import EasyDict
from pygame.math import Vector2

from .base_manager import BaseManager
from gobigger.utils import format_vector, Border
from gobigger.balls import FoodBall, ThornsBall, CloneBall, SporeBall


class FoodManager(BaseManager):

    def __init__(self, cfg, border, random_generator=None):
        super(FoodManager, self).__init__(cfg, border)
        self.food_refresh_time = self.cfg.refresh_time
        self.refresh_time_count = 0
        if random_generator is not None:
            self._random = random_generator
        else:
            self._random = random.Random()

    def get_balls(self):
        return list(self.balls.values())
    
    def add_balls(self, balls):
        if isinstance(balls, list):
            for ball in balls:
                self.balls[ball.name] = ball
        elif isinstance(balls, FoodBall):
            self.balls[balls.name] = balls
        return True

    def refresh(self):
        todo_num = min(self.cfg.refresh_num, self.cfg.num_max - len(self.balls))
        for _ in range(todo_num):
            self.add_balls(self.spawn_ball())

    def remove_balls(self, balls):
        if isinstance(balls, list):
            for ball in balls:
                ball.remove()
                try:
                    del self.balls[ball.name]
                except:
                    pass
        elif isinstance(balls, FoodBall):
            balls.remove()
            try:
                del self.balls[balls.name]
            except:
                pass

    def spawn_ball(self, position=None, size=None):
        if position is None:
            position = self.border.sample()
        if size is None:
            size = self._random.uniform(self.ball_settings.radius_min, self.ball_settings.radius_max)**2
        name = uuid.uuid1()
        return FoodBall(name=name, position=position, border=self.border, size=size, **self.ball_settings)

    def init_balls(self, custom_init=None):
        if custom_init is None:
            for _ in range(self.cfg.num_init):
                ball = self.spawn_ball()
                self.balls[ball.name] = ball
        else:
            for ball_cfg in custom_init:
                ball = self.spawn_ball(position=Vector2(*ball_cfg[:2]), size=ball_cfg[2]**2)
                self.balls[ball.name] = ball

    def step(self, duration):
        self.refresh_time_count += duration
        if self.refresh_time_count >= self.food_refresh_time:
            self.refresh()
            self.refresh_time_count = 0

    def reset(self):
        self.refresh_time_count = 0
        self.balls = {}
        return True



import uuid
from pygame.math import Vector2
import logging

from .base_player import BasePlayer
from gobigger.balls import FoodBall, ThornsBall, CloneBall, SporeBall


class HumanPlayer(BasePlayer):

    def __init__(self, cfg, team_name, name, border, spore_settings):
        self.team_name = team_name
        self.name = name
        self.border = border
        self.balls = {}
        self.ball_settings = cfg
        self.spore_settings = spore_settings
        self.stop_flag = False

    def get_clone_num(self):
        '''
        Overview:
            Get how many avatars the current player has
        '''
        return len(self.balls)

    def get_balls(self):
        '''
        Overview:
            Get all the balls of the current player
        '''
        return list(self.balls.values())

    def add_balls(self, balls):
        '''
        Overview:
            Add new avatars
        Parameters:
            balls <List[CloneBall] or CloneBall>: It can be a list or a single doppelganger
        '''
        if isinstance(balls, list):
            for ball in balls:
                self.balls[ball.name] = ball
        elif isinstance(balls, CloneBall):
            self.balls[balls.name] = balls
        return True

    def move(self, direction=None, duration=0.05):
        '''
        Overview:
            Move all balls controlled by the player
            The main logic is
             1. Processing stopped state
             2. If it is stopping, control all balls to move closer to the center of mass
        Parameters:
            direction <Vector2>: A point in the unit circle
            duration <float>: time
        Returns:
            position <Vector2>: position after moving 
        '''
        if direction is not None: 
            self.stop_flag = False # Exit stopped state
            for ball in self.balls.values():
                ball.stop_flag = False
        if self.get_clone_num() == 0:
            pass
        elif self.stop_flag: 
            if self.get_clone_num() == 1: 
                for ball in self.balls.values():
                    ball.move(given_acc=None, given_acc_center=None, duration=duration)
            else: #If there are multiple balls, control them to move to the center
                centroid = self.cal_centroid()
                for ball in self.balls.values():
                    given_acc_center = centroid - ball.position
                    if given_acc_center.length() == 0:
                        given_acc_center = Vector2(0.000001, 0.000001)
                    elif given_acc_center.length() > 1:
                        given_acc_center = given_acc_center.normalize()
                    given_acc_center *= 20
                    ball.move(given_acc=direction, given_acc_center=given_acc_center, duration=duration)
        else:
            if self.get_clone_num() == 1:
                for ball in self.balls.values():
                    ball.move(given_acc=direction, given_acc_center=Vector2(0,0), duration=duration)
            else:
                centroid = self.cal_centroid()
                for ball in self.balls.values():
                    given_acc_center = centroid - ball.position
                    if given_acc_center.length() == 0:
                        given_acc_center = Vector2(0.000001, 0.000001)
                    elif given_acc_center.length() > 1:
                        given_acc_center = given_acc_center.normalize()
                    given_acc_center *= 20
                    ball.move(given_acc=direction, given_acc_center=given_acc_center, duration=duration)
        self.size_decay()
        return True

    def size_decay(self):
        '''
        Overview: 
            The player’s balls' size will decay over time
        '''
        for ball in self.balls.values():
            ball.size_decay()
        return True

    def eject(self, direction=None):
        '''
        Overview:
            All clones controlled by the player perform the spore-spitting action
        Return:
            <list>: list of new spores
        '''
        ret = []
        ball_names = list(self.balls.keys())
        for ball_name in ball_names:
            if ball_name in self.balls:
                ball = self.balls[ball_name]
                ret.append(ball.eject(direction=direction))
        return ret

    def get_keys_sort_by_balls(self):
        '''
        Overview:
            Sort by ball size from largest to smallest
        Return:
            <list>: list of names
        '''
        items = self.balls.items() 
        backitems=[[v[1],v[0]] for v in items] 
        backitems.sort(reverse=True) 
        return [ backitems[i][1] for i in range(0,len(backitems))] 

    def split(self, direction=None):
        '''
        Overview:
            All avatars controlled by the player perform splits, from large to small
        '''
        balls_keys = self.get_keys_sort_by_balls()
        for k in balls_keys:
            if k in self.balls:
                ret = self.balls[k].split(self.get_clone_num(), direction=direction)
                if ret and isinstance(ret, CloneBall):
                    self.add_balls(ret)
        return True

    def eat(self, ball):
        raise NotImplementedError

    def stop(self):
        if self.stop_flag:
            return True
        self.stop_flag = True
        centroid = self.cal_centroid()
        for ball in self.balls.values():
            direction_center = centroid - ball.position
            if direction_center.length() == 0:
                direction_center = Vector2(0.000001, 0.000001)
            ball.stop(direction=direction_center)
        return True

    def remove_balls(self, ball):
        ball.remove()
        if ball.name in self.balls:
            try:
                del self.balls[ball.name]
            except:
                pass
        return True

    def respawn(self, position):
        ball_name = uuid.uuid1()
        ball = CloneBall(team_name=self.team_name, name=ball_name, position=position, border=self.border, owner=self.name, 
                         spore_settings=self.spore_settings, **self.ball_settings)
        direction = Vector2(1, 0)
        ball.stop()
        self.balls = {}
        self.balls[ball.name] = ball   
        return True

    def cal_centroid(self):
        '''
        Overview:
            Calculate the centroid
        '''
        x = 0
        y = 0
        total_size = 0
        for ball in self.get_balls():
            x += ball.size * ball.position.x
            y += ball.size * ball.position.y
            total_size += ball.size
        return Vector2(x, y) / total_size

    def adjust(self):
        '''
        Overview:
            Adjust all the balls controlled by the player, including two parts
            1. Possible Rigid Body Collision
            2. Possible ball-ball fusion
        '''
        balls = self.get_balls()
        balls = sorted(balls, reverse=True)
        balls_num = len(balls)
        to_remove_balls = []
        for i in range(balls_num-1):
            if not balls[i].is_remove:
                for j in range(i+1, balls_num):
                    if not balls[j].is_remove:
                        dis = balls[i].get_dis(balls[j])
                        if dis < balls[i].radius + balls[j].radius:
                            if balls[i].judge_rigid(balls[j]):
                                balls[i].rigid_collision(balls[j]) # Rigid body collision                       
                            else:
                                if dis < balls[i].radius or dis < balls[j].radius: 
                                    if balls[i].size > balls[j].size:
                                        balls[i].eat(balls[j])
                                        balls[j].remove()
                                        to_remove_balls.append(balls[j])
                                    else:
                                        balls[j].eat(balls[i])
                                        balls[i].remove()
                                        to_remove_balls.append(balls[i])
        for ball in to_remove_balls: 
            self.remove_balls(ball)

    def get_total_size(self):
        '''
            Overview: 
                Get the total size of all balls of the current player
        '''
        total_size = 0
        for ball in self.get_balls():
            total_size += ball.size
        return total_size









server_default_config = dict(
    team_num=4, 
    player_num_per_team=3, 
    map_width=1000,
    map_height=1000, 
    match_time=60*10,
    state_tick_per_second=10, # frame
    action_tick_per_second=5, # frame
    collision_detection_type='precision',
    save_video=False,
    save_quality='high', # ['high', 'low']
    save_path='',
    save_bin=False, # save bin to go-explore
    load_bin=False,
    load_bin_path='',
    load_bin_frame_num = 'all',
    jump_to_frame_file = '',
    manager_settings=dict(
        # food setting
        food_manager=dict(
            num_init=2000, # initial number
            num_min=2000, # Minimum number
            num_max=2500, # Maximum number
            refresh_time=2, # Time interval (seconds) for refreshing food in the map
            refresh_num=30, # The number of refreshed foods in the map each time
            ball_settings=dict( # The specific parameter description can be viewed in the ball module
                radius_min=2,
                radius_max=2,
            ),
        ),
        # thorns setting
        thorns_manager=dict(
            num_init=15, # initial number
            num_min=15, # Minimum number
            num_max=20, # Maximum number
            refresh_time=2, # Time interval (seconds) for refreshing thorns in the map
            refresh_num=2, # The number of refreshed  thorns in the map each time
            ball_settings=dict( # The specific parameter description can be viewed in the ball module
                radius_min=12, 
                radius_max=20, 
                vel_max=100,
                eat_spore_vel_init=10, 
                eat_spore_vel_zero_time=1,
            )
        ),
        # player setting
        player_manager=dict(
            ball_settings=dict(  # The specific parameter description can be viewed in the ball module
                acc_max=100, 
                vel_max=25,
                radius_min=3, 
                radius_max=300, 
                radius_init=3, 
                part_num_max=16, 
                on_thorns_part_num=10, 
                on_thorns_part_radius_max=20,
                split_radius_min=10, 
                eject_radius_min=10, 
                recombine_age=20,
                split_vel_init=30,
                split_vel_zero_time=1, 
                stop_zero_time=1,
                size_decay_rate=0.00005, 
                given_acc_weight=10,
            )
        ),
        # spore setting
        spore_manager=dict(
            ball_settings=dict( # The specific parameter description can be viewed in the ball module
                radius_min=3, 
                radius_max=3, 
                vel_init=250,
                vel_zero_time=0.3, 
                spore_radius_init=20, 
            )
        )   
    ),
    custom_init=dict(
        food=[], # only position and radius
        thorns=[], # only position and radius
        spore=[], # only position and radius
        clone=[], # only position and radius and player and team
    ),
    obs_settings=dict(
        with_spatial=True,
        with_speed=False,
        with_all_vision=False,
        cheat=False,
    ),
)



import logging
import pytest
import uuid
from pygame.math import Vector2
import pygame
import numpy as np
import cv2
import argparse
import time
import importlib

from gobigger.utils import Border
from gobigger.server import Server
from gobigger.render import RealtimeRender, RealtimePartialRender, EnvRender
from gobigger.agents import BotAgent

logging.basicConfig(level=logging.DEBUG)


def play_by_config(config_name):
    config_module = importlib.import_module('gobigger.hyper.configs.config_{}'.format(config_name))
    config = config_module.server_default_config
    server = Server(config)
    server.reset()
    render = RealtimeRender(server.map_width, server.map_height)
    server.set_render(render)
    human_team_name = '0'
    human_team_player_name = []
    bot_agents = []
    for player in server.player_manager.get_players():
        if player.team_name != human_team_name:
            bot_agents.append(BotAgent(player.name))
        else:
            human_team_player_name.append(player.name)
    fps_real = 0
    t1 = time.time()
    clock = pygame.time.Clock()
    fps_set = server.state_tick_per_second
    for i in range(100000):
        obs = server.obs()
        # actions_bot = {bot_agent.name: bot_agent.step(obs[1][bot_agent.name]) for bot_agent in bot_agents}
        actions_bot = {bot_agent.name: [None, None, -1] for bot_agent in bot_agents}
        actions = {player_name: [None, None, -1] for player_name in human_team_player_name}
        x, y = None, None
        action_type = -1
        # ================ control by keyboard ===============
        for event in pygame.event.get():
            if event.type == pygame.KEYDOWN:
                x1, y1, x2, y2 = None, None, None, None
                action_type1 = -1
                action_type2 = -1
                if event.key == pygame.K_UP:
                    x1, y1 = 0, -1
                if event.key == pygame.K_DOWN:
                    x1, y1 = 0, 1
                if event.key == pygame.K_LEFT:
                    x1, y1 = -1, 0
                if event.key == pygame.K_RIGHT:
                    x1, y1 = 1, 0
                if event.key == pygame.K_LEFTBRACKET: # Spores
                    action_type1 = 0
                if event.key == pygame.K_RIGHTBRACKET: # Splite
                    action_type1 = 1
                if event.key == pygame.K_BACKSLASH: # Stop moving
                    action_type1 = 2
                if event.key == pygame.K_w:
                    x2, y2 = 0, -1
                if event.key == pygame.K_s:
                    x2, y2 = 0, 1
                if event.key == pygame.K_a:
                    x2, y2 = -1, 0
                if event.key == pygame.K_d:
                    x2, y2 = 1, 0
                if event.key == pygame.K_1: # Spores
                    action_type2 = 0
                if event.key == pygame.K_2: # Splite
                    action_type2 = 1
                if event.key == pygame.K_3: # Stop moving
                    action_type2 = 2
                actions = {
                    human_team_player_name[0]: [x1, y1, action_type1],
                    human_team_player_name[1]: [x2, y2, action_type2],
                }
        if server.last_time < server.match_time:
            actions.update(actions_bot)
            print(actions)
            server.step_state_tick(actions=actions)
            if actions is not None and x is not None and y is not None:
                render.fill(server, direction=Vector2(x, y), fps=fps_real, last_time=server.last_time)
            else:
                render.fill(server, direction=None, fps=fps_real, last_time=server.last_time)
            render.show()
            if i % server.state_tick_per_second == 0:
                t2 = time.time()
                fps_real = server.state_tick_per_second/(t2-t1)
                t1 = time.time()
        else:
            logging.debug('Game Over')
            break
        clock.tick(fps_set)
    render.close()


if __name__ == '__main__':
    parser = argparse.ArgumentParser()
    parser.add_argument('-c', '--config', type=str, default='2f2s')
    args = parser.parse_args()

    play_by_config(args.config)
    


import os
import sys
import time
import pygame
from pygame.math import Vector2
import queue
import math
from easydict import EasyDict

from gobigger.server import Server
from gobigger.render import EnvRender, RealtimeRender


class HyperAction:
    def __init__(self):
        raise NotImplementedError
    def update(self, obs):
        raise NotImplementedError
    def get(self):
        raise NotImplementedError


class StraightMergeHyperAction(HyperAction):

    #########################################################
    #                        直线中合                        #
    #########################################################

    def __init__(self, player_name1, player_name2):
        self.player_name1 = player_name1
        self.player_name2 = player_name2
        self.state = "reach" # in ['reach', 'merge']
        self.need_stop = 0

    def update(self, obs1, obs2):
        self.balls1 = self.get_balls_from_obs_by_name(self.player_name1, obs1)
        self.balls2 = self.get_balls_from_obs_by_name(self.player_name2, obs2)

    def get(self):
        if self.state == 'reach':
            tmp = self.straight_merge_reach()
            if tmp is not None:
                return tmp
        if self.state == 'action':
            return self.straight_merge_action()

    def get_balls_from_obs_by_name(self, player_name, obs):
        balls = []
        overlap = obs['overlap']
        for item in overlap['clone']:
            if str(int(item[3])) == player_name:
                ball = EasyDict()
                ball.position = Vector2(float(item[0]), float(item[1]))
                ball.radius = float(item[2])
                ball.player = str(int(item[3]))
                ball.team = str(int(item[4]))
                balls.append(ball)
        return balls

    def straight_merge_reach(self):
        """
        移动到相近位置（相切）
        """
        ball1 = self.balls1[0]
        ball2 = self.balls2[0]
        if len(self.balls1) == 1 and len(self.balls2) == 1 and \
            (ball1.position - ball2.position).length() <= 1.5 * (ball1.radius + ball2.radius):
            self.state = 'action'
            return None
        if len(self.balls1) != 1:
            action1 = [None, None, 2]
        else:
            direction1 = ball2.position - ball1.position
            action1 = [direction1.x, direction1.y, -1]
        if len(self.balls2) != 1:
            action2 = [None, None, 2]
        else:
            direction2 = ball1.position - ball2.position
            action2 = [direction2.x, direction2.y, -1]
        actions = {
            str(ball1.player): action1,
            str(ball2.player): action2,
        }
        return actions

    def straight_merge_action(self):
        """
        直线中合
        balls1是较大的球，balls2是较小的球
        条件：1. 同队友都只有一个球。
        """
        ret = {}
        ball1 = self.balls1[0]
        ball2 = self.balls2[0]
        if self.need_stop > 0:
            self.need_stop -= 1
            return {
                ball1.player: [None, None, 2],
                ball2.player: [None, None, 2],
            }
        size1 = ball1.radius ** 2
        size2 = ball2.radius ** 2
        split_size_min = 100
        if len(self.balls1) == 1:
            tmp = ball2.position - ball1.position
            direction = [tmp.x, tmp.y]
        else:
            direction = [None, None]
        if size1 > split_size_min:
            ret[ball1.player] = [*direction, 1] # split
            ret[ball2.player] = [None, None, 2] # stop
            self.need_stop = 5
        else:
            ret = None
        return ret


class QuarterMergeHyperAction(HyperAction):

    #########################################################
    #                        四分中合                        #
    #########################################################

    def __init__(self, player_name1, player_name2):
        self.player_name1 = player_name1
        self.player_name2 = player_name2
        self.state = "reach" # in ['reach', 'merge']
        self.need_stop = 0
        self.split_count = 0

    def update(self, obs1, obs2):
        self.balls1 = self.get_balls_from_obs_by_name(self.player_name1, obs1)
        self.balls2 = self.get_balls_from_obs_by_name(self.player_name2, obs2)

    def get(self):
        if self.state == 'reach':
            tmp = self.quarter_merge_reach()
            if tmp is not None:
                return tmp
        if self.state == 'action':
            return self.quarter_merge_action()

    def get_balls_from_obs_by_name(self, player_name, obs):
        balls = []
        overlap = obs['overlap']
        for item in overlap['clone']:
            if str(int(item[3])) == player_name:
                ball = EasyDict()
                ball.position = Vector2(float(item[0]), float(item[1]))
                ball.radius = float(item[2])
                ball.player = str(int(item[3]))
                ball.team = str(int(item[4]))
                balls.append(ball)
        return balls

    def quarter_merge_reach(self):
        """
        移动到相近位置（相切）
        """
        ball1 = self.balls1[0]
        ball2 = self.balls2[0]
        if len(self.balls1) == 1 and len(self.balls2) == 1 and \
            (ball1.position - ball2.position).length() <= 1.5 * (ball1.radius + ball2.radius):
            self.state = 'action'
            return None
        if len(self.balls1) != 1:
            action1 = [None, None, 2]
        else:
            direction1 = ball2.position - ball1.position
            action1 = [direction1.x, direction1.y, -1]
        if len(self.balls2) != 1:
            action2 = [None, None, 2]
        else:
            direction2 = ball1.position - ball2.position
            action2 = [direction2.x, direction2.y, -1]
        actions = {
            str(ball1.player): action1,
            str(ball2.player): action2,
        }
        return actions

    def quarter_merge_action(self):
        """
        直线中合
        balls1是较大的球，balls2是较小的球
        条件：1. 同队友都只有一个球。
        """
        def cal_centroid(balls):
            '''
            Overview:
                Calculate the centroid
            '''
            x = 0
            y = 0
            total_size = 0
            for ball in balls:
                x += ball.radius ** 2 * ball.position.x
                y += ball.radius ** 2 * ball.position.y
                total_size += ball.radius ** 2
            return Vector2(x, y) / total_size

        ret = {}
        ball1 = self.balls1[0]
        ball2 = self.balls2[0]
        if self.need_stop > 0:
            self.need_stop -= 1
            return {
                ball1.player: [None, None, 2],
                ball2.player: [None, None, 2],
            }
        size1 = ball1.radius ** 2
        size2 = ball2.radius ** 2
        split_size_min = 100
        centroid = cal_centroid(self.balls1)
        if len(self.balls1) == 1 or self.split_count == 0:
            tmp = ball2.position - ball1.position
            tmp = Vector2(tmp.x * math.cos(math.pi/4) - tmp.y * math.sin(math.pi/4),
                          tmp.x * math.sin(math.pi/4) + tmp.y * math.cos(math.pi/4))
            direction = [tmp.x, tmp.y]
            self.split_count += 1
        elif self.split_count == 1:
            tmp = ball2.position - centroid
            direction = [tmp.x, tmp.y]
            self.split_count += 1
        else:
            direction = [None, None]
            self.split_count += 1
        if size1 > split_size_min:
            ret[ball1.player] = [*direction, 1] # split
            tmp = centroid - ball2.position
            ret[ball2.player] = [tmp.x, tmp.y, 2] # stop
            self.need_stop = 5
        else:
            ret = None
        return ret


class EighthMergeHyperAction(HyperAction):

    #########################################################
    #                        四分中合                        #
    #########################################################

    def __init__(self, player_name1, player_name2):
        self.player_name1 = player_name1
        self.player_name2 = player_name2
        self.state = "reach" # in ['reach', 'merge']
        self.need_stop = 0
        self.split_count = 0

    def update(self, obs1, obs2):
        self.balls1 = self.get_balls_from_obs_by_name(self.player_name1, obs1)
        self.balls2 = self.get_balls_from_obs_by_name(self.player_name2, obs2)

    def get(self):
        if self.state == 'reach':
            tmp = self.quarter_merge_reach()
            if tmp is not None:
                return tmp
        if self.state == 'action':
            return self.quarter_merge_action()

    def get_balls_from_obs_by_name(self, player_name, obs):
        balls = []
        overlap = obs['overlap']
        for item in overlap['clone']:
            if str(int(item[3])) == player_name:
                ball = EasyDict()
                ball.position = Vector2(float(item[0]), float(item[1]))
                ball.radius = float(item[2])
                ball.player = str(int(item[3]))
                ball.team = str(int(item[4]))
                balls.append(ball)
        return balls

    def quarter_merge_reach(self):
        """
        移动到相近位置（相切）
        """
        ball1 = self.balls1[0]
        ball2 = self.balls2[0]
        if len(self.balls1) == 1 and len(self.balls2) == 1 and \
            (ball1.position - ball2.position).length() <= 1.5 * (ball1.radius + ball2.radius):
            self.state = 'action'
            return None
        if len(self.balls1) != 1:
            action1 = [None, None, 2]
        else:
            direction1 = ball2.position - ball1.position
            action1 = [direction1.x, direction1.y, -1]
        if len(self.balls2) != 1:
            action2 = [None, None, 2]
        else:
            direction2 = ball1.position - ball2.position
            action2 = [direction2.x, direction2.y, -1]
        actions = {
            str(ball1.player): action1,
            str(ball2.player): action2,
        }
        return actions

    def quarter_merge_action(self):
        """
        直线中合
        balls1是较大的球，balls2是较小的球
        条件：1. 同队友都只有一个球。
        """
        def cal_centroid(balls):
            '''
            Overview:
                Calculate the centroid
            '''
            x = 0
            y = 0
            total_size = 0
            for ball in balls:
                x += ball.radius ** 2 * ball.position.x
                y += ball.radius ** 2 * ball.position.y
                total_size += ball.radius ** 2
            return Vector2(x, y) / total_size

        ret = {}
        ball1 = self.balls1[0]
        ball2 = self.balls2[0]
        if self.need_stop > 0:
            self.need_stop -= 1
            return {
                ball1.player: [None, None, 2],
                ball2.player: [None, None, 2],
            }
        size1 = ball1.radius ** 2
        size2 = ball2.radius ** 2
        split_size_min = 100
        centroid = cal_centroid(self.balls1)
        if len(self.balls1) == 1 or self.split_count == 0:
            tmp = ball2.position - ball1.position
            tmp = Vector2(tmp.x * math.cos(math.pi/4) - tmp.y * math.sin(math.pi/4),
                          tmp.x * math.sin(math.pi/4) + tmp.y * math.cos(math.pi/4))
            direction = [tmp.x, tmp.y]
            self.split_count += 1
        elif self.split_count == 1:
            tmp = ball2.position - centroid
            direction = [tmp.x, tmp.y]
            self.split_count += 1
        else:
            direction = [None, None]
            self.split_count += 1
        if size1 > split_size_min:
            ret[ball1.player] = [*direction, 1] # split
            tmp = centroid - ball2.position
            ret[ball2.player] = [tmp.x, tmp.y, 2] # stop
            self.need_stop = 5
        else:
            tmp = ball2.position - centroid
            ret = {
                ball1.player: [tmp.x, tmp.y, -1],
                ball2.player: [None, None, 2],
            }
        return ret





from gobigger.balls import FoodBall, ThornsBall, CloneBall, SporeBall


class BasePlayer:
    '''
    Player's abstract class
    '''
    def __init__(self, name=None):
        self.name = name

    def move(self, direction):
        '''
        Parameters:
            direction <Vector2>: Given any point in a unit circle, the angle represents the direction, and the magnitude represents the acceleration
        '''
        raise NotImplementedError

    def eject(self):
        '''
        Do sporulation
        '''
        raise NotImplementedError

    def eat(self, ball):
        '''
        Eat another ball
        '''
        raise NotImplementedError

    def stop(self):
        '''
        stop moving
        '''
        raise NotImplementedError

    def respawn(self):
        raise NotImplementedError


import logging
import pytest
import uuid
from pygame.math import Vector2
import time
import numpy as np
import cv2
import pygame
import pickle

from gobigger.agents import BotAgent
from gobigger.utils import Border
from gobigger.server import Server
from gobigger.render import RealtimeRender, RealtimePartialRender, EnvRender

logging.basicConfig(level=logging.DEBUG)


def demo_bot():
    server = Server(dict(
        team_num=4, 
        player_num_per_team=3, 
        map_width=1000, 
        map_height=1000, 
        match_time=60*10,
        state_tick_per_second=10, # frame
        action_tick_per_second=5, 
        collision_detection_type='precision',
        save_video=True,
        save_quality='low', # ['high', 'low']
        save_path='',
        save_bin=False, # save bin to go-explore
        load_bin=False,
        load_bin_path='',
        jump_to_frame_file='',
        manager_settings=dict(
            # food setting
            food_manager=dict(
                num_init=2000, # initial number
                num_min=2000, # Minimum number
                num_max=2500, # Maximum number
                refresh_time=2, # Time interval (seconds) for refreshing food in the map
                refresh_num=30, # The number of refreshed foods in the map each time
                ball_settings=dict( # The specific parameter description can be viewed in the ball module
                    radius_min=2, 
                    radius_max=2,
                ),
            ),
            # thorns setting
            thorns_manager=dict(
                num_init=15, # initial number
                num_min=15, # Minimum number
                num_max=20, # Maximum number
                refresh_time=2, # Time interval (seconds) for refreshing thorns in the map
                refresh_num=2, # The number of refreshed  thorns in the map each time
                ball_settings=dict( # The specific parameter description can be viewed in the ball module
                    radius_min=12, 
                    radius_max=20, 
                    vel_max=100,
                    eat_spore_vel_init=10, 
                    eat_spore_vel_zero_time=1,
                )
            ),
            # player setting
            player_manager=dict(
                ball_settings=dict(  # The specific parameter description can be viewed in the ball module
                    acc_max=100, 
                    vel_max=25,
                    radius_min=3, 
                    radius_max=100, 
                    radius_init=3, 
                    part_num_max=16, 
                    on_thorns_part_num=10, 
                    on_thorns_part_radius_max=20,
                    split_radius_min=10, 
                    eject_radius_min=10, 
                    recombine_age=20,
                    split_vel_init=30,
                    split_vel_zero_time=1, 
                    stop_zero_time=1,
                    size_decay_rate=0.00005, 
                    given_acc_weight=10,
                )
            ),
            # spore setting
            spore_manager=dict(
                ball_settings=dict( # The specific parameter description can be viewed in the ball module
                    radius_min=3, 
                    radius_max=3, 
                    vel_init=250,
                    vel_zero_time=0.3, 
                    spore_radius_init=20, 
                )
            )
        ),
        custom_init=dict(
            food=[],
            thorns=[],
            spore=[],
            clone=[],
        ),
    ))
    render = EnvRender(server.map_width, server.map_height)
    server.set_render(render)
    server.reset()
    bot_agents = []
    for player in server.player_manager.get_players():
        #bot_agents.append(LowAgent(player.name))
        #bot_agents.append(MidAgent(player.name))
        bot_agents.append(BotAgent(player.name, level=1))
    time_obs = 0
    time_step = 0
    time_fill_all = 0
    time_get_rectangle = 0
    time_get_clip = 0
    time_cvt = 0
    time_overlap = 0
    for i in range(100000):
        t1 = time.time()
        obs = server.obs()
        t2 = time.time()
        actions = {bot_agent.name: bot_agent.step(obs[1][bot_agent.name]) for bot_agent in bot_agents}
        t3 = time.time()
        finish_flag = server.step(actions=actions)
        t4 = time.time()
        tmp_obs = t2-t1
        tmp_step = t4-t3
        time_obs +=  tmp_obs
        time_step += tmp_step
        logging.debug('{} {:.4f} obs: {:.3f} / {:.3f}, step: {:.3f} / {:.3f}'\
            .format(i, server.last_time, tmp_obs, time_obs/(i+1), tmp_step, time_step/(i+1)))
        if finish_flag:
            logging.debug('Game Over')
            break
    server.close()


if __name__ == '__main__':
    demo_bot()



import logging
import pytest
import uuid
from pygame.math import Vector2
import pygame
import cv2
import math

from .base_render import BaseRender
from .env_render import EnvRender
from gobigger.utils import FOOD_COLOR, THORNS_COLOR, SPORE_COLOR, PLAYER_COLORS, BACKGROUND, RED, WHITE
from gobigger.utils import to_aliased_circle


class RealtimeRender(EnvRender):
    '''
    Overview:
        Used in real-time games, giving a global view
    '''
    def __init__(self, width, height, padding=(0,0), cell_size=10, only_render=False):
        super(RealtimeRender, self).__init__(width, height, padding=padding, 
                                             cell_size=cell_size, only_render=only_render)

    def fill(self, server, direction=None, fps=0, last_time=0, player_num_per_team=3):
        self.screen.fill(BACKGROUND)
        # Render all balls
        self.screen, _ = self.render_all_balls_colorful(self.screen, 
                                                        food_balls=server.food_manager.get_balls(), 
                                                        thorns_balls=server.thorns_manager.get_balls(), 
                                                        spore_balls=server.spore_manager.get_balls(), 
                                                        players=server.player_manager.get_players(), 
                                                        player_num_per_team=player_num_per_team)
        # for debug
        font= pygame.font.SysFont('Menlo', 15, True)

        team_name_size = {}
        for player in server.player_manager.get_players():
            if player.team_name not in team_name_size:
                team_name_size[player.team_name] = player.get_total_size()
            else:
                team_name_size[player.team_name] += player.get_total_size()
        team_name_size = sorted(team_name_size.items(), key=lambda d: d[1], reverse=True)
        for index, (team_name, team_size) in enumerate(team_name_size):
            pos_txt= font.render('{}: {:.5f}'.format(team_name, team_size), 1, RED)
            self.screen.blit(pos_txt, (20, 10+10*(index*2+1)))

        fps_txt = font.render(
            'fps: ' + str(fps), 1, RED)
        last_time_txt = font.render(
            'last_time: ' + str(last_time), 1, RED)
        self.screen.blit(fps_txt, (20, self.height_full - 30))
        self.screen.blit(last_time_txt, (20, self.height_full - 50))

    
    def show(self):
        pygame.display.update()
        self.fpsClock.tick(self.FPS)

    def close(self):
        pygame.quit()


class RealtimePartialRender(EnvRender):
    '''
    Overview:
        Used in real-time games to give the player a visible field of view. The corresponding player can be obtained by specifying the player name. The default is the first player
    '''
    def __init__(self, width, height, padding=(0,0), cell_size=10, 
                 scale_up_ratio=1.5, vision_x_min=100, vision_y_min=100, player_name=None, only_render=False):
        super(RealtimePartialRender, self).__init__(width, height, padding=padding, 
                                                    cell_size=cell_size, scale_up_ratio=scale_up_ratio, 
                                                    vision_x_min=vision_x_min, vision_y_min=vision_y_min,
                                                    only_render=only_render)
        self.player_name = player_name

    def fill(self, server, direction=None, fps=0, last_time=0, player_num_per_team=3):
        if self.player_name is None:
            player = server.player_manager.get_players()[0]
        else:
            player = server.player_manager.get_player_by_name(self.player_name)
        rectangle = self.get_rectangle_by_player(player)

        screen = pygame.Surface((self.width, self.height))
        screen.fill(BACKGROUND)

        screen, _ = self.render_all_balls_colorful(screen, 
                                                   food_balls=server.food_manager.get_balls(), 
                                                   thorns_balls=server.thorns_manager.get_balls(), 
                                                   spore_balls=server.spore_manager.get_balls(), 
                                                   players=server.player_manager.get_players(), 
                                                   player_num_per_team=1)

        screen_data = pygame.surfarray.array3d(screen)

        screen_data = self.get_clip_screen(screen_data, rectangle)
        player_surface = pygame.pixelcopy.make_surface(screen_data)
        screen_data = pygame.transform.scale(player_surface, (self.width, self.height))
        self.screen.blit(screen_data, (0, 0))

        # for debug
        font= pygame.font.SysFont('Menlo', 15, True)

        team_name_size = {}
        for player in server.player_manager.get_players():
            if player.team_name not in team_name_size:
                team_name_size[player.team_name] = player.get_total_size()
            else:
                team_name_size[player.team_name] += player.get_total_size()
        team_name_size = sorted(team_name_size.items(), key=lambda d: d[1], reverse=True)
        for index, (team_name, team_size) in enumerate(team_name_size):
            pos_txt= font.render('{}: {:.5f}'.format(team_name, team_size), 1, RED)
            self.screen.blit(pos_txt, (20, 10+10*(index*2+1)))

        fps_txt = font.render('fps: ' + str(fps), 1, RED)
        last_time_txt = font.render('last_time: ' + str(last_time), 1, RED)
        self.screen.blit(fps_txt, (20, self.height_full - 30))
        self.screen.blit(last_time_txt, (20, self.height_full - 50))
    
    def show(self):
        pygame.display.update()
        self.fpsClock.tick(self.FPS)

    def close(self):
        pygame.quit()



import random
from easydict import EasyDict
import uuid
import logging
import cv2
import os
import sys
import time
import numpy as np
import copy
import pickle

os.environ['PYGAME_HIDE_SUPPORT_PROMPT'] = "hide"
os.environ['SDL_AUDIODRIVER'] = 'dsp'

from pygame.math import Vector2

from gobigger.utils import Border, create_collision_detection, deep_merge_dicts
from gobigger.balls import FoodBall, ThornsBall, CloneBall, SporeBall
from gobigger.managers import FoodManager, SporeManager, ThornsManager, PlayerManager
from .server_default_config import server_default_config


class Server:
    '''
    Overview:
        Server is responsible for the management of the entire game environment, including the status of all balls in the environment, and the status update after the action is entered
        The main logic when updating is as follows:
        0 tick -> input action -> update the state of the player's ball -> update the state of all balls after the current state continues for 1 tick
            -> detect collision and eating (update status) -> 0 tick end/1 tick start
        The details are as follows:
        1. Generate all balls (food, thorns, players)
        2. Single step
            1. Modify the current state of all players' balls according to the action (including acceleration, instantaneous state after splitting/spitting)
            2. Continue a tick for the current state of all balls, that is, update the acceleration/velocity/position of each ball after a tick, and at the same time the ball that is in the moving state in this tick
            3. Adjust all balls in each player (rigid body collision + ball-ball fusion)
            4. For each ball moved in this tick (already sorted by priority):
                1. We will know which balls he collided with (there will be repetitions)
                2. Category discussion
                    1. One of the balls is the player ball
                        1. Another is another player's ball, the bigger one eats the smaller one
                        2. The other side is your own clone, in fact, you don’t need to deal with it if you have already dealt with it before.
                        3. Another is food/spores, player ball eat it
                        4. Another is the thornball
                            1. Do not touch the center of the circle, continue
                            2. Hit the center of the circle
                                1. player ball is older than thornball
                                    1. number of player's avatar reaches the upper limit, thornball will be eaten
                                    2. number of player's avatar doesn't reache the upper limit, player ball eat thornball and blow up
                                2. player ball is younger than thornball, nothing happened
                    2. One of the balls is a thornball
                        1. Another is the player ball
                            1. hit the center of a circle
                                1. player ball is older than thornball
                                    1. number of player's avatar reaches the upper limit, thornball will be eaten
                                    2. number of player's avatar doesn't reache the upper limit, player ball eat thornball and blow up
                                2. player ball is younger than thornball, nothing happened
                            2. Do not touch the center of the circle, continue,
                        2. The another is the spore, thornball eat it and add a speed and acceleration
                    3. One of the balls is Spore
                        1. Another is the player ball, Spore was eaten
                        2. Another is the thorn ball, Spore was eaten
        3. After each tick, check if you want to update food, thorns, and player rebirth
    '''

    @staticmethod
    def default_config():
        cfg = copy.deepcopy(server_default_config)
        return EasyDict(cfg)

    def __init__(self, cfg=None, seed=None):
        self.cfg = Server.default_config()
        if isinstance(cfg, dict):
            cfg = EasyDict(cfg)
            self.cfg = deep_merge_dicts(self.cfg, cfg)
        logging.debug(self.cfg)
        self.team_num = self.cfg.team_num
        self.player_num_per_team = self.cfg.player_num_per_team
        self.map_width = self.cfg.map_width
        self.map_height = self.cfg.map_height
        self.match_time = self.cfg.match_time
        self.state_tick_per_second = self.cfg.state_tick_per_second
        self.action_tick_per_second = self.cfg.action_tick_per_second
        # other kwargs
        self.state_tick_duration = 1 / self.state_tick_per_second
        self.action_tick_duration = 1 / self.action_tick_per_second
        self.state_tick_per_action_tick = self.state_tick_per_second // self.action_tick_per_second

        self.custom_init = self.cfg.custom_init
        self.jump_to_frame_file = self.cfg.jump_to_frame_file
        self.save_video = self.cfg.save_video
        self.save_quality = self.cfg.save_quality
        self.save_path = self.cfg.save_path
        self.save_bin = self.cfg.save_bin
        self.load_bin = self.cfg.load_bin
        self.load_bin_path = self.cfg.load_bin_path
        self.load_bin_frame_num = self.cfg.load_bin_frame_num
        self.obs_settings = self.cfg.obs_settings
        
        self.seed(seed)
        self.border = Border(0, 0, self.map_width, self.map_height, self._random)
        self.last_time = 0
        self.screens_all = []
        self.screens_partial = {}
        self.actions_record = []

        self.food_manager = FoodManager(self.cfg.manager_settings.food_manager, border=self.border, random_generator=self._random)
        self.thorns_manager = ThornsManager(self.cfg.manager_settings.thorns_manager, border=self.border, random_generator=self._random)
        self.spore_manager = SporeManager(self.cfg.manager_settings.spore_manager, border=self.border, random_generator=self._random)
        self.player_manager  = PlayerManager(self.cfg.manager_settings.player_manager, border=self.border,
                                             team_num=self.team_num, player_num_per_team=self.player_num_per_team, 
                                             spore_manager_settings=self.cfg.manager_settings.spore_manager,
                                             random_generator=self._random)

        self.collision_detection_type = self.cfg.collision_detection_type
        self.collision_detection = create_collision_detection(self.collision_detection_type, border=self.border)

    def spawn_balls(self):
        '''
        Overview:
            Initialize all balls. If self.custom_init is set, initialize all balls based on it.
        '''
        # check custom_init
        self.custom_init_food = self.custom_init.food if self.custom_init.food else None
        self.custom_init_thorns = self.custom_init.thorns if self.custom_init.thorns else None
        self.custom_init_spore = self.custom_init.spore if self.custom_init.spore else None
        self.custom_init_clone = self.custom_init.clone if self.custom_init.clone else None
        self.load_frame_info()
        # init
        self.food_manager.init_balls(custom_init=self.custom_init_food) # init food
        self.thorns_manager.init_balls(custom_init=self.custom_init_thorns) # init thorns
        self.spore_manager.init_balls(custom_init=self.custom_init_spore) # init spore
        self.player_manager.init_balls(custom_init=self.custom_init_clone) # init player

    def step_state_tick(self, actions=None):
        moving_balls = [] # Record all balls in motion
        total_balls = [] # Record all balls
        # Update all player balls according to action
        if actions is not None:
            '''
            In a single action: 
              If sporulation and splitting operations occur at the same time, sporulation will be given priority
              If move and stop move occur at the same time in action, perform stop move operation
            '''
            for player in self.player_manager.get_players():
                direction_x, direction_y, action_type = actions[player.name]
                if direction_x is None or direction_y is None:
                    direction = None
                else:
                    direction = Vector2(direction_x, direction_y)
                    if direction.length() == 0:
                        direction = None
                    else:
                        direction = direction.normalize()
                if action_type == 0 or action_type == 3: # eject
                    tmp_spore_balls = player.eject(direction=direction)
                    for tmp_spore_ball in tmp_spore_balls:
                        if tmp_spore_ball:
                            self.spore_manager.add_balls(tmp_spore_ball) 
                if action_type == 1 or action_type == 4: # split 
                    self.player_manager.add_balls(player.split(direction=direction))
                if action_type == 2: # stop moving
                    player.stop()
                elif action_type == 3 or action_type == 4: # move on old direction
                    player.move(duration=self.state_tick_duration)
                    moving_balls.extend(player.get_balls())
                else: # move on new direction
                    player.move(direction=direction, duration=self.state_tick_duration)
                    moving_balls.extend(player.get_balls())
        else:
            for player in self.player_manager.get_players():
                player.move(duration=self.state_tick_duration)
                moving_balls.extend(player.get_balls())
                total_balls.extend(player.get_balls())
        moving_balls = sorted(moving_balls, reverse=True) # Sort by size
        # Update the status of other balls after moving, and record the balls with status updates
        for thorns_ball in self.thorns_manager.get_balls():
            if thorns_ball.moving:
                thorns_ball.move(duration=self.state_tick_duration)
            moving_balls.append(thorns_ball)
        for spore_ball in self.spore_manager.get_balls():
            if spore_ball.moving:
                spore_ball.move(duration=self.state_tick_duration)
        # Adjust the position of all player balls
        self.player_manager.adjust()
        # Collision detection
        total_balls.extend(self.player_manager.get_balls())
        total_balls.extend(self.thorns_manager.get_balls())
        total_balls.extend(self.spore_manager.get_balls())
        total_balls.extend(self.food_manager.get_balls())
        collisions_dict = self.collision_detection.solve(moving_balls, total_balls)
        # Process each ball in moving_balls
        for index, moving_ball in enumerate(moving_balls):
            if not moving_ball.is_remove and index in collisions_dict:
                for target_ball in collisions_dict[index]:
                    self.deal_with_collision(moving_ball, target_ball)
        # After each tick, check if there is a need to update food, thorns, and player rebirth
        self.food_manager.step(duration=self.state_tick_duration)
        self.spore_manager.step(duration=self.state_tick_duration)
        self.thorns_manager.step(duration=self.state_tick_duration)
        self.player_manager.step()
        self.last_time += self.state_tick_duration

    def deal_with_collision(self, moving_ball, target_ball):
        if not moving_ball.is_remove and not target_ball.is_remove: # Ensure that the two balls are present
            if isinstance(moving_ball, CloneBall): 
                if isinstance(target_ball, CloneBall):
                    if moving_ball.team_name != target_ball.team_name:
                        if moving_ball.size > target_ball.size:
                            moving_ball.eat(target_ball)
                            self.player_manager.remove_balls(target_ball)
                        else:
                            target_ball.eat(moving_ball)
                            self.player_manager.remove_balls(moving_ball)
                    elif moving_ball.owner != target_ball.owner:
                        if moving_ball.size > target_ball.size:
                            if self.player_manager.get_clone_num(target_ball) > 1:
                                moving_ball.eat(target_ball)
                                self.player_manager.remove_balls(target_ball)
                        else:
                            if self.player_manager.get_clone_num(moving_ball) > 1:
                                target_ball.eat(moving_ball)
                                self.player_manager.remove_balls(moving_ball)
                elif isinstance(target_ball, FoodBall):
                    moving_ball.eat(target_ball)
                    self.food_manager.remove_balls(target_ball)
                elif isinstance(target_ball, SporeBall):
                    moving_ball.eat(target_ball)
                    self.spore_manager.remove_balls(target_ball)
                elif isinstance(target_ball, ThornsBall):
                    if moving_ball.size > target_ball.size:
                        ret = moving_ball.eat(target_ball, clone_num=self.player_manager.get_clone_num(moving_ball))
                        self.thorns_manager.remove_balls(target_ball)
                        if isinstance(ret, list): 
                            self.player_manager.add_balls(ret) 
            elif isinstance(moving_ball, ThornsBall):
                if isinstance(target_ball, CloneBall):
                    if moving_ball.size < target_ball.size: 
                        ret = target_ball.eat(moving_ball, clone_num=self.player_manager.get_clone_num(target_ball))
                        self.thorns_manager.remove_balls(moving_ball)
                        if isinstance(ret, list): 
                            self.player_manager.add_balls(ret) 
                elif isinstance(target_ball, SporeBall): 
                    moving_ball.eat(target_ball)
                    self.spore_manager.remove_balls(target_ball)
            elif isinstance(moving_ball, SporeBall):
                if isinstance(target_ball, CloneBall) or isinstance(target_ball, ThornsBall): 
                    target_ball.eat(moving_ball)
                    self.spore_manager.remove_balls(moving_ball)
        else:
            return

    def start(self):
        self.spawn_balls()
        self._end_flag = False

    def stop(self):
        self._end_flag = True

    def reset(self):
        self.last_time = 0
        self.screens_all = []
        self.screens_partial = {}
        self.actions_record = []
        self.load_record()
        self.food_manager.reset()
        self.thorns_manager.reset()
        self.spore_manager.reset()
        self.player_manager.reset()
        self.start()
        self.resume_actions()

    def record_frame_for_video(self):
        if self.save_video:
            screen_data_all, screen_data_players = self.render.get_tick_all_colorful(
                food_balls=self.food_manager.get_balls(),
                thorns_balls=self.thorns_manager.get_balls(),
                spore_balls=self.spore_manager.get_balls(),
                players=self.player_manager.get_players(),
                player_num_per_team=self.player_num_per_team,
                team_name_size=self.player_manager.get_teams_size())
            self.screens_all.append(screen_data_all)
            for player_name, screen_data_player in screen_data_players.items():
                if player_name not in self.screens_partial:
                    self.screens_partial[player_name] = []
                self.screens_partial[player_name].append(screen_data_player)

    def step(self, actions=None, save_frame_full_path='', **kwargs):
        self.save_frame_info(save_frame_full_path=save_frame_full_path)
        if self.last_time >= self.match_time:
            if self.save_video:
                self.save_mp4(save_path=self.save_path)
            if self.save_bin:
                self.save_record(save_path=self.save_path)
            self.stop()
            return True
        if not self._end_flag:
            self.actions_record.append(actions)
            for i in range(self.state_tick_per_action_tick):
                if i == 0:
                    self.step_state_tick(actions)
                    self.record_frame_for_video()
                else:
                    self.step_state_tick()
                    if self.save_quality == 'high':
                        self.record_frame_for_video()
        return False

    def set_render(self, render):
        self.render = render
        self.render.set_obs_settings(self.obs_settings)

    def obs(self, obs_type='all'):
        assert obs_type in ['all', 'single']
        assert hasattr(self, 'render')
        team_name_size = self.player_manager.get_teams_size()
        global_state = {
            'border': [self.map_width, self.map_height],
            'total_time': self.match_time,
            'last_time': self.last_time,
            'leaderboard': {
                str(i): team_name_size[str(i)] for i in range(self.team_num)
            }
        }
        _, screen_data_players = self.render.update_all(food_balls=self.food_manager.get_balls(),
                                                        thorns_balls=self.thorns_manager.get_balls(),
                                                        spore_balls=self.spore_manager.get_balls(),
                                                        players=self.player_manager.get_players())
        return global_state, screen_data_players

    def save_mp4(self, save_path=''):
        # self.video_id = time.strftime("%Y-%m-%d-%H-%M-%S", time.localtime())
        self.video_id = str(uuid.uuid1())
        fps = self.state_tick_per_second if self.save_quality == 'high' else self.action_tick_per_second
        # save all
        video_file = os.path.join(save_path, '{}-all.mp4'.format(self.video_id))
        out = cv2.VideoWriter(video_file, cv2.VideoWriter_fourcc(*'mp4v'), fps, (self.screens_all[0].shape[1], self.screens_all[0].shape[0]))
        for screen in self.screens_all:
            out.write(screen)
        out.release()
        cv2.destroyAllWindows()
        # save partial
        for player_name, screens in self.screens_partial.items():
            video_file = os.path.join(save_path, '{}-{:02d}.mp4'.format(self.video_id, int(player_name)))
            out = cv2.VideoWriter(video_file, cv2.VideoWriter_fourcc(*'mp4v'), fps, screens[0].shape[:2])
            for screen in screens:
                out.write(screen)
            out.release()
            cv2.destroyAllWindows()

    def save_record(self, save_path=''):
        data = {'seed': self._seed, 'actions': self.actions_record}
        with open(os.path.join(save_path, '{}.pkl'.format(str(uuid.uuid1()))), 'wb') as f:
            pickle.dump(data, f)

    def load_record(self):
        if self.load_bin and os.path.isfile(self.load_bin_path):
            with open(self.load_bin_path, 'rb') as f:
                data = pickle.load(f)
            seed = data['seed']
            self.actions_record_last = data['actions']
            self.seed(seed=seed)

    def resume_actions(self):
        if self.load_bin and os.path.isfile(self.load_bin_path):
            if self.load_bin_frame_num == 'all':
                self.load_bin_frame_num = len(self.actions_record_last)
            for action in self.actions_record_last[:self.load_bin_frame_num]:
                self.step(action)

    def save_frame_info(self, save_frame_full_path):
        if save_frame_full_path != '':
            frame_info = {'food': [], 'thorns': [], 'spore': [], 'clone': []}
            # food
            for ball in self.food_manager.get_balls():
                frame_info['food'].append([ball.position.x, ball.position.y, ball.radius])
            # thorns
            for ball in self.thorns_manager.get_balls():
                frame_info['thorns'].append([ball.position.x, ball.position.y, ball.radius, ball.vel.x, ball.vel.y,
                                           ball.acc.x, ball.acc.y, ball.move_time, ball.moving])
            # spore
            for ball in self.spore_manager.get_balls():
                frame_info['spore'].append([ball.position.x, ball.position.y, ball.radius, ball.direction.x, 
                                           ball.direction.y, ball.vel.x, ball.vel.y,
                                           ball.acc.x, ball.acc.y, ball.move_time, ball.moving])
            # clone
            for ball in self.player_manager.get_balls():
                frame_info['clone'].append([ball.position.x, ball.position.y, ball.radius, ball.owner, 
                                           ball.team_name, ball.vel.x, ball.vel.y, ball.acc.x, ball.acc.y, 
                                           ball.vel_last.x, ball.vel_last.y, ball.acc_last.x, ball.acc_last.y, 
                                           ball.direction.x, ball.direction.y, ball.last_given_acc.x, 
                                           ball.last_given_acc.y, ball.age, ball.cooling_last, ball.stop_flag,
                                           ball.stop_time, ball.acc_stop.x, ball.acc_stop.y])
            with open(save_frame_full_path, 'wb') as f:
                pickle.dump(frame_info, f)

    def load_frame_info(self):
        if self.jump_to_frame_file:
            with open(self.jump_to_frame_file, 'rb') as f:
                data = pickle.load(f)
            self.custom_init_food = data['food']
            self.custom_init_thorns = data['thorns']
            self.custom_init_spore = data['spore']
            self.custom_init_clone = data['clone']

    def get_player_names(self):
        return self.player_manager.get_player_names()

    def get_team_names(self):
        return self.player_manager.get_team_names()

    def get_player_names_with_team(self):
        return self.player_manager.get_player_names_with_team()

    def close(self):
        self.stop()
        if hasattr(self, 'render'):
            self.render.close()

    def seed(self, seed=None):
        if seed is None:
            self._seed = random.randrange(sys.maxsize)
        else:
            self._seed = seed
        self._random = random.Random(self._seed)



import math
import logging
from abc import ABC, abstractmethod
from easydict import EasyDict
from pygame.math import Vector2

from gobigger.utils import format_vector, Border
from gobigger.balls import FoodBall, ThornsBall, CloneBall, SporeBall


class BaseManager(ABC):
    '''
    Overview:
        Base class for all ball managers
    '''
    def __init__(self, cfg, border):
        self.cfg = cfg
        self.border = border
        self.balls = {}
        self.ball_settings = self.cfg.ball_settings

    def get_balls(self):
        '''
        Overview:
            Get all balls currently managed
        '''
        return self.balls.values()
    
    def add_balls(self, balls):
        '''
        Overview:
            Add one (or more) balls
        '''
        raise NotImplementedError

    def refresh(self):
        '''
        Overview:
            Refresh. Used to refresh the balls in management. Such as replenishing eaten food balls
        '''
        raise NotImplementedError

    def remove_balls(self, balls):
        '''
        Overview:
            Remove managed balls
        '''
        raise NotImplementedError

    def spawn_ball(self):
        raise NotImplementedError

    def init_balls(self):
        raise NotImplementedError

    def step(self, duration):
        '''
        Overview:
            Perform a status update under the control of the server
        '''
        raise NotImplementedError

    def obs(self):
        '''
        Overview:
            Return data available for observation
        '''
        raise NotImplementedError

    def reset(self):
        raise NotImplementedError



import logging
from easydict import EasyDict

from gobigger.utils import format_vector, add_size, Border, deep_merge_dicts
from .base_ball import BaseBall


class FoodBall(BaseBall):
    """
    Overview:
        - characteristic:
        * Can't move, can only be eaten, randomly generated
    """
    @staticmethod
    def default_config():
        cfg = BaseBall.default_config()
        cfg.update(dict())
        return EasyDict(cfg)

    def __init__(self, name, position, border, size=1, vel=None, acc=None, **kwargs):
        super(FoodBall, self).__init__(name, position, border, size=size, vel=vel, acc=acc, **kwargs)
        self.check_border()

    def move(self, direction, duration):
        logging.debug('FoodBall can not move')
        return

    def eat(self, ball):
        logging.debug('FoodBall can not eat others')
        return



from .structures import format_vector, add_size, save_screen_data_to_img, Border, QuadNode
from .collision_detection import create_collision_detection
from .config_utils import deep_merge_dicts
from .tool import *
from .colors import *



import logging
import pytest
import uuid
from pygame.math import Vector2
import pygame
import numpy as np
import cv2
import argparse
import time

from gobigger.utils import Border
from gobigger.server import Server
from gobigger.render import RealtimeRender, RealtimePartialRender, EnvRender
from gobigger.agents import BotAgent

logging.basicConfig(level=logging.DEBUG)


def play_control_by_keyboard():
    server = Server(dict(
            team_num=1, 
            player_num_per_team=1,
            match_time=60*10,
            state_tick_per_second=20, # frame
            action_tick_per_second=5, # frame
        ))
    server.start()
    render = RealtimeRender(server.map_width, server.map_height)
    fps_real = 0
    t1 = time.time()
    clock = pygame.time.Clock()
    fps_set = server.state_tick_per_second
    for i in range(100000):
        clock.tick(1000)
        actions = None
        x, y = None, None
        action_type = -1
        # ================ control by keyboard ===============
        for event in pygame.event.get():
            if event.type == pygame.KEYDOWN:
                if event.key == pygame.K_UP:
                    x, y = 0, -1
                    action_type = -1
                elif event.key == pygame.K_DOWN:
                    x, y = 0, 1
                    action_type = -1
                elif event.key == pygame.K_LEFT:
                    x, y = -1, 0
                    action_type = -1
                elif event.key == pygame.K_RIGHT:
                    x, y = 1, 0
                    action_type = -1
                elif event.key == pygame.K_q: # Spores
                    x, y = None, None
                    action_type = 0
                elif event.key == pygame.K_w: # Splite
                    x, y = None, None
                    action_type = 1
                elif event.key == pygame.K_e: # Stop moving
                    x, y = None, None
                    action_type = 2
                elif event.key == pygame.K_r: # Cheating, adding weight
                    x, y = None, None
                    action_type = -1
                    server.player_manager.get_players()[0].get_balls()[0].set_size(144)
                actions = {player.name: [x, y, action_type] for player in server.player_manager.get_players()}
        if server.last_time < server.match_time:
            server.step_state_tick(actions=actions)
            if actions is not None and x is not None and y is not None:
                render.fill(server, direction=Vector2(x, y), fps=fps_real, last_time=server.last_time)
            else:
                render.fill(server, direction=None, fps=fps_real, last_time=server.last_time)
            render.show()
            if i % server.state_tick_per_second == 0:
                t2 = time.time()
                fps_real = server.state_tick_per_second/(t2-t1)
                t1 = time.time()
        else:
            logging.debug('Game Over')
            break
        clock.tick(fps_set)
    render.close()

def play_control_by_keyboard_multi_player():
    server = Server(dict(
            team_num=2,
            player_num_per_team=1, 
            match_time=60*10,
            state_tick_per_second=20, # frame
            action_tick_per_second=5, # frame
        ))
    server.start()
    render = RealtimeRender(server.map_width, server.map_height)
    player_names = [player.name for player in server.player_manager.get_players()]
    fps_real = 0
    t1 = time.time()
    clock = pygame.time.Clock()
    fps_set = server.state_tick_per_second
    for i in range(100000):
        actions = None
        # ================ control by keyboard ===============
        for event in pygame.event.get():
            if event.type == pygame.KEYDOWN:
                x1, y1, x2, y2 = None, None, None, None
                action_type1 = -1
                action_type2 = -1
                if event.key == pygame.K_UP:
                    x1, y1 = 0, -1
                if event.key == pygame.K_DOWN:
                    x1, y1 = 0, 1
                if event.key == pygame.K_LEFT:
                    x1, y1 = -1, 0
                if event.key == pygame.K_RIGHT:
                    x1, y1 = 1, 0
                if event.key == pygame.K_LEFTBRACKET: # Spores
                    action_type1 = 0
                if event.key == pygame.K_RIGHTBRACKET: # Splite
                    action_type1 = 1
                if event.key == pygame.K_BACKSLASH: # Stop moving
                    action_type1 = 2
                if event.key == pygame.K_w:
                    x2, y2 = 0, -1
                if event.key == pygame.K_s:
                    x2, y2 = 0, 1
                if event.key == pygame.K_a:
                    x2, y2 = -1, 0
                if event.key == pygame.K_d:
                    x2, y2 = 1, 0
                if event.key == pygame.K_1: # Spores
                    action_type2 = 0
                if event.key == pygame.K_2: # Splite
                    action_type2 = 1
                if event.key == pygame.K_3: # Stop moving
                    action_type2 = 2
                actions = {
                    player_names[0]: [x1, y1, action_type1],
                    player_names[1]: [x2, y2, action_type2],
                }
        if server.last_time < server.match_time:
            server.step_state_tick(actions=actions)
            if actions is not None and x1 is not None and y1 is not None:
                render.fill(server, direction=Vector2(x1, y1), fps=fps_real, last_time=server.last_time)
            else:
                render.fill(server, direction=None, fps=fps_real, last_time=server.last_time)
            render.show()
            if i % server.state_tick_per_second == 0:
                t2 = time.time()
                fps_real = server.state_tick_per_second/(t2-t1)
                t1 = time.time()
        else:
            logging.debug('Game Over')
            break
        clock.tick(fps_set)
    render.close()

def play_control_by_keyboard_partial():
    '''
    Keyboard control + partly visible
    '''
    server = Server(dict(
            team_num=1, 
            player_num_per_team=1, 
            match_time=60*10, 
            state_tick_per_second=20, # frame
            action_tick_per_second=5, # frame
        ))
    server.start()
    render = RealtimePartialRender(server.map_width, server.map_height)
    fps_real = 0
    t1 = time.time()
    clock = pygame.time.Clock()
    fps_set = server.state_tick_per_second
    for i in range(100000):
        actions = None
        x, y = None, None
        action_type = -1
        # ================ control by keyboard ===============
        for event in pygame.event.get():
            if event.type == pygame.KEYDOWN:
                if event.key == pygame.K_UP:
                    x, y = 0, -1
                    action_type = -1
                elif event.key == pygame.K_DOWN:
                    x, y = 0, 1
                    action_type = -1
                elif event.key == pygame.K_LEFT:
                    x, y = -1, 0
                    action_type = -1
                elif event.key == pygame.K_RIGHT:
                    x, y = 1, 0
                    action_type = -1
                elif event.key == pygame.K_q: # Spore
                    x, y = None, None
                    action_type = 0
                elif event.key == pygame.K_w: # Splite
                    x, y = None, None
                    action_type = 1
                elif event.key == pygame.K_e: # stop moving
                    x, y = None, None
                    action_type = 2
                elif event.key == pygame.K_r: # cheating
                    x, y = None, None
                    action_type = -1
                    server.player_manager.get_players()[0].get_balls()[0].set_size(144)
                actions = {player.name: [x, y, action_type] for player in server.player_manager.get_players()}

        if server.last_time < server.match_time:
            server.step_state_tick(actions=actions)
            if actions is not None and x is not None and y is not None:
                render.fill(server, direction=Vector2(x, y), fps=fps_real, last_time=server.last_time)
            else:
                render.fill(server, direction=None, fps=fps_real, last_time=server.last_time)
            render.show()
            if i % server.state_tick_per_second == 0:
                t2 = time.time()
                fps_real = server.state_tick_per_second/(t2-t1)
                t1 = time.time()
        else:
            logging.debug('Game Over')
            break
        clock.tick(fps_set)
    render.close()

def play_control_by_keyboard_vs_bot(team_num=2):
    server = Server(dict(
            team_num=team_num, 
            player_num_per_team=1,
            match_time=60*10,
            save_video=True,
            state_tick_per_second=20, # frame
            action_tick_per_second=5, # frame
        ))
    server.start()
    render = RealtimeRender(server.map_width, server.map_height)
    server.set_render(render)
    human_team_name = '0'
    human_team_player_name = []
    bot_agents = []
    for player in server.player_manager.get_players():
        if player.team_name != human_team_name:
            bot_agents.append(BotAgent(player.name))
        else:
            human_team_player_name.append(player.name)
    fps_real = 0
    t1 = time.time()
    clock = pygame.time.Clock()
    fps_set = server.state_tick_per_second
    for i in range(100000):
        obs = server.obs()
        actions_bot = {bot_agent.name: bot_agent.step(obs[1][bot_agent.name]) for bot_agent in bot_agents}
        actions = {player_name: [None, None, -1] for player_name in human_team_player_name}
        x, y = None, None
        action_type = -1
        # ================ control by keyboard ===============
        for event in pygame.event.get():
            if event.type == pygame.KEYDOWN:
                if event.key == pygame.K_UP:
                    x, y = 0, -1
                    action_type = -1
                elif event.key == pygame.K_DOWN:
                    x, y = 0, 1
                    action_type = -1
                elif event.key == pygame.K_LEFT:
                    x, y = -1, 0
                    action_type = -1
                elif event.key == pygame.K_RIGHT:
                    x, y = 1, 0
                    action_type = -1
                elif event.key == pygame.K_q: # Spores
                    x, y = None, None
                    action_type = 0
                elif event.key == pygame.K_w: # Splite
                    x, y = None, None
                    action_type = 1
                elif event.key == pygame.K_e: # Stop moving
                    x, y = None, None
                    action_type = 2
                elif event.key == pygame.K_r: # Cheating, adding weight
                    x, y = None, None
                    action_type = -1
                    server.player_manager.get_players()[0].get_balls()[0].set_size(144)
                actions = {player_name: [x, y, action_type] for player_name in human_team_player_name}
        if server.last_time < server.match_time:
            actions.update(actions_bot)
            server.step_state_tick(actions=actions)
            if actions is not None and x is not None and y is not None:
                render.fill(server, direction=Vector2(x, y), fps=fps_real, last_time=server.last_time)
            else:
                render.fill(server, direction=None, fps=fps_real, last_time=server.last_time)
            render.show()
            if i % server.state_tick_per_second == 0:
                t2 = time.time()
                fps_real = server.state_tick_per_second/(t2-t1)
                t1 = time.time()
        else:
            logging.debug('Game Over')
            break
        clock.tick(fps_set)
    render.close()


def play_by_bot(team_num, player_num, agent_class):

    class TeamAgent:
        def __init__(self, team_name, player_names):
            self.agents = {}
            self.team_name = team_name
            self.player_names = player_names
            for player_name in self.player_names:
                self.agents[player_name] = BotAgent(name=player_name)

        def get_actions(self, obs):
            global_state, player_states = obs
            actions = {}
            for player_name, agent in self.agents.items():
                action = agent.step(player_states[player_name])
                actions[player_name] = action
            return actions

    server = Server(dict(
        team_num=team_num,
        player_num_per_team=player_num,  # 每个队伍的玩家数量
        match_time=60 * 10,  # 每场比赛的持续时间
        save_video=False,
        state_tick_per_second=20, # frame
        action_tick_per_second=5, # frame
    ))
    render = RealtimeRender(server.map_width, server.map_height)
    server.set_render(render)
    server.start()
    agents = []
    team_player_names = server.get_team_names()
    team_names = list(team_player_names.keys())
    for index in range(server.team_num):
        try:
            if agent_class == '':
                p = TeamAgent
            else:
                from importlib import import_module
                class_path, class_name = agent_class.split(':')
                p = import_module(class_path)
                p = getattr(p, class_name)
            agents.append(p(team_name=team_names[index],
                                         player_names=team_player_names[team_names[index]]))
        except Exception as e:
            print('load bot class failed.')
            exit()

    for i in range(60 * 10 * server.action_tick_per_second + 1):
        obs = server.obs()
        render.show()
        event = pygame.event.poll()
        render.fill(server, direction=None, fps=60, last_time=server.last_time)
        global_state, player_states = obs
        actions = {}
        for agent in agents:
            agent_obs = [global_state, {
                player_name: player_states[player_name] for player_name in agent.player_names
            }]
            actions.update(agent.get_actions(agent_obs))
        finish_flag = server.step(actions=actions)
        if finish_flag:
            logging.debug('Game Over!')
            break
    server.close()


if __name__ == '__main__':
    parser = argparse.ArgumentParser()
    parser.add_argument('--player-num', type=int, choices=[1,2,3], default=1)
    parser.add_argument('--vision-type', type=str, choices=['full', 'partial'], default='full')
    parser.add_argument('--vs-bot', action='store_true', help='vs bot')
    parser.add_argument('--bot-only', action='store_true', help='bot only')
    parser.add_argument('--team-num', type=int, default=2)
    parser.add_argument('--agent-class', type=str, default='', help='agent class, like \''
                                                                    'submit.bot_submission:BotSubmission\'')
    args = parser.parse_args()

    if args.bot_only:
        play_by_bot(team_num=args.team_num, player_num=args.player_num, agent_class=args.agent_class)
    elif args.vs_bot:
        play_control_by_keyboard_vs_bot(team_num=args.team_num)
    elif args.player_num == 1 and args.vision_type == 'full':
        play_control_by_keyboard()
    elif args.player_num == 1 and args.vision_type == 'partial':
        play_control_by_keyboard_partial()
    elif args.player_num == 2 and args.vision_type == 'full':
        play_control_by_keyboard_multi_player()
    else:
        logging.error('Not supoort when player_num = {} and vision_type = {}'.format(args.player_num, args.vision_type))



import logging
from easydict import EasyDict
from pygame.math import Vector2

from gobigger.utils import format_vector, add_size, Border, deep_merge_dicts
from .base_ball import BaseBall
from .spore_ball import SporeBall


class ThornsBall(BaseBall):
    """
    Overview:
        - characteristic:
        * Can't move actively
        * Can eat spores. When eating spores, it will inherit the momentum of the spores and move a certain distance.
        * Can only be eaten by balls heavier than him. After eating, it will split the host into multiple smaller units.
        * Nothing happens when a ball lighter than him passes by
    """
    @staticmethod
    def default_config():
        cfg = BaseBall.default_config()
        cfg.update(dict(
            radius_min=5, # Minimum radius, greater than the player's maximum number of clones multiplied by the player's minimum weight
            radius_max=10, # Maximum radius
            vel_max=100, # Maximum velocity
            eat_spore_vel_init=100, # Initialization speed after eating spores
            eat_spore_vel_zero_time=0.2, # Time to zero speed after eating spores(s)
        ))
        return EasyDict(cfg)

    def __init__(self, name, position, border, size=1, vel=None, acc=None, **kwargs):
        # init other kwargs
        kwargs = EasyDict(kwargs)
        cfg = ThornsBall.default_config()
        cfg = deep_merge_dicts(cfg, kwargs)
        super(ThornsBall, self).__init__(name, position, border, size=size, vel=vel, acc=acc, **cfg)
        self.vel_max = cfg.vel_max
        self.radius_min = cfg.radius_min
        self.radius_max = cfg.radius_max
        self.eat_spore_vel_init = cfg.eat_spore_vel_init
        self.eat_spore_vel_zero_time = cfg.eat_spore_vel_zero_time
        self.acc_default = self.eat_spore_vel_init / self.eat_spore_vel_zero_time
        self.move_time = 0
        self.moving = False
        self.check_border()

    def move(self, direction=None, duration=0.05):
        assert direction is None
        assert duration > 0
        self.position = self.position + self.vel * duration
        self.move_time += duration
        if self.move_time < self.eat_spore_vel_zero_time:
            self.vel += self.acc * duration
        else:
            self.vel = Vector2(0, 0)
            self.acc = Vector2(0, 0)
            self.moving = False
        self.check_border()
        return True

    def eat(self, ball):
        if isinstance(ball, SporeBall):
            self.set_size(add_size(self.size, ball.size))
            if self.radius > self.radius_max:
                self.radius = self.radius_max
            if ball.vel.length() > 0:
                self.vel = self.eat_spore_vel_init * ball.vel.normalize()
                self.acc = - self.acc_default * ball.vel.normalize()
                self.move_time = 0
                self.moving = True
        else:
            logging.debug('ThornsBall can not eat {}'.format(type(ball)))
        return True



from .base_manager import BaseManager
from .food_manager import FoodManager
from .spore_manager import SporeManager
from .thorns_manager import ThornsManager
from .player_manager import PlayerManager



import os
import random
import logging
import copy
import queue
from pygame.math import Vector2

from .base_agent import BaseAgent


class BotAgent(BaseAgent):
    '''
    Overview:
        A simple script bot
    '''
    def __init__(self, name=None, level=3):
        self.name = name
        self.actions_queue = queue.Queue()
        self.last_clone_num = 1
        self.last_total_size = 0
        self.level = level

    def step(self, obs):
        if self.level == 1:
            return self.step_level_1(obs)
        if self.level == 2:
            return self.step_level_2(obs)
        if self.level == 3:
            return self.step_level_3(obs)

    def step_level_1(self, obs):
        if self.actions_queue.qsize() > 0:
            return self.actions_queue.get()
        overlap = obs['overlap']
        overlap = self.preprocess(overlap)
        food_balls = overlap['food']
        thorns_balls = overlap['thorns']
        spore_balls = overlap['spore']
        clone_balls = overlap['clone']

        my_clone_balls, others_clone_balls = self.process_clone_balls(clone_balls)
        min_distance, min_food_ball = self.process_food_balls(food_balls, my_clone_balls[0])        
        if min_food_ball is not None:
            direction = (min_food_ball['position'] - my_clone_balls[0]['position']).normalize()
        else:
            direction = (Vector2(0, 0) - my_clone_balls[0]['position']).normalize()
        action_type = -1
        self.actions_queue.put([direction.x, direction.y, action_type])
        action_ret = self.actions_queue.get()
        return action_ret

    def step_level_2(self, obs):
        if self.actions_queue.qsize() > 0:
            return self.actions_queue.get()
        overlap = obs['overlap']
        overlap = self.preprocess(overlap)
        food_balls = overlap['food']
        thorns_balls = overlap['thorns']
        spore_balls = overlap['spore']
        clone_balls = overlap['clone']

        my_clone_balls, others_clone_balls = self.process_clone_balls(clone_balls)
        min_distance, min_thorns_ball = self.process_thorns_balls(thorns_balls, my_clone_balls[0])
        if min_thorns_ball is not None:
            direction = (min_thorns_ball['position'] - my_clone_balls[0]['position']).normalize()
        else:
            direction = (Vector2(0, 0) - my_clone_balls[0]['position']).normalize()
        action_type = -1
        self.actions_queue.put([direction.x, direction.y, action_type])
        self.actions_queue.put([None, None, -1])
        self.actions_queue.put([None, None, -1])
        self.actions_queue.put([None, None, -1])
        self.actions_queue.put([None, None, -1])
        self.actions_queue.put([None, None, -1])
        action_ret = self.actions_queue.get()
        return action_ret

    def step_level_3(self, obs):
        if self.actions_queue.qsize() > 0:
            return self.actions_queue.get()
        overlap = obs['overlap']
        overlap = self.preprocess(overlap)
        food_balls = overlap['food']
        thorns_balls = overlap['thorns']
        spore_balls = overlap['spore']
        clone_balls = overlap['clone']

        my_clone_balls, others_clone_balls = self.process_clone_balls(clone_balls)

        if len(my_clone_balls) >= 9 and my_clone_balls[4]['radius'] > 14:
            self.actions_queue.put([None, None, 2])
            self.actions_queue.put([None, None, -1])
            self.actions_queue.put([None, None, -1])
            self.actions_queue.put([None, None, -1])
            self.actions_queue.put([None, None, -1])
            self.actions_queue.put([None, None, -1])
            self.actions_queue.put([None, None, -1])
            self.actions_queue.put([None, None, 0])
            self.actions_queue.put([None, None, 0])
            self.actions_queue.put([None, None, 0])
            self.actions_queue.put([None, None, 0])
            self.actions_queue.put([None, None, 0])
            self.actions_queue.put([None, None, 0])
            self.actions_queue.put([None, None, 0])
            self.actions_queue.put([None, None, 0])
            action_ret = self.actions_queue.get()
            return action_ret

        if len(others_clone_balls) > 0 and my_clone_balls[0]['radius'] < others_clone_balls[0]['radius']:
            direction = (my_clone_balls[0]['position'] - others_clone_balls[0]['position']).normalize()
            action_type = -1
        else:
            min_distance, min_thorns_ball = self.process_thorns_balls(thorns_balls, my_clone_balls[0])
            if min_thorns_ball is not None:
                direction = (min_thorns_ball['position'] - my_clone_balls[0]['position']).normalize()
            else:
                min_distance, min_food_ball = self.process_food_balls(food_balls, my_clone_balls[0])
                if min_food_ball is not None:
                    direction = (min_food_ball['position'] - my_clone_balls[0]['position']).normalize()
                else:
                    direction = (Vector2(0, 0) - my_clone_balls[0]['position']).normalize()
            action_random = random.random()
            if action_random < 0.02:
                action_type = 1
            if action_random < 0.04 and action_random > 0.02:
                action_type = 2
            else:
                action_type = -1
        direction = self.add_noise_to_direction(direction)
        self.actions_queue.put([direction.x, direction.y, action_type])
        action_ret = self.actions_queue.get()
        return action_ret

    def process_clone_balls(self, clone_balls):
        my_clone_balls = []
        others_clone_balls = []
        for clone_ball in clone_balls:
            if clone_ball['player'] == self.name:
                my_clone_balls.append(copy.deepcopy(clone_ball))
        my_clone_balls.sort(key=lambda a: a['radius'], reverse=True)

        for clone_ball in clone_balls:
            if clone_ball['player'] != self.name:
                others_clone_balls.append(copy.deepcopy(clone_ball))
        others_clone_balls.sort(key=lambda a: a['radius'], reverse=True)
        return my_clone_balls, others_clone_balls

    def process_thorns_balls(self, thorns_balls, my_max_clone_ball):
        min_distance = 10000
        min_thorns_ball = None
        for thorns_ball in thorns_balls:
            if thorns_ball['radius'] < my_max_clone_ball['radius']:
                distance = (thorns_ball['position'] - my_max_clone_ball['position']).length()
                if distance < min_distance:
                    min_distance = distance
                    min_thorns_ball = copy.deepcopy(thorns_ball)
        return min_distance, min_thorns_ball

    def process_food_balls(self, food_balls, my_max_clone_ball):
        min_distance = 10000
        min_food_ball = None
        for food_ball in food_balls:
            distance = (food_ball['position'] - my_max_clone_ball['position']).length()
            if distance < min_distance:
                min_distance = distance
                min_food_ball = copy.deepcopy(food_ball)
        return min_distance, min_food_ball

    def preprocess(self, overlap):
        new_overlap = {}
        for k, v in overlap.items():
            if k =='clone':
                new_overlap[k] = []
                for index, vv in enumerate(v):
                    tmp={}
                    tmp['position'] = Vector2(vv[0],vv[1])
                    tmp['radius'] = vv[2]
                    tmp['player'] = str(int(vv[-2]))
                    tmp['team'] = str(int(vv[-1]))
                    new_overlap[k].append(tmp)
            else:
                new_overlap[k] = []
                for index, vv in enumerate(v):
                    tmp={}
                    tmp['position'] = Vector2(vv[0],vv[1])
                    tmp['radius'] = vv[2]
                    new_overlap[k].append(tmp)
        return new_overlap

    def preprocess_tuple2vector(self, overlap):
        new_overlap = {}
        for k, v in overlap.items():
            new_overlap[k] = []
            for index, vv in enumerate(v):
                new_overlap[k].append(vv)
                new_overlap[k][index]['position'] = Vector2(*vv['position'])
        return new_overlap
    
    def add_noise_to_direction(self, direction, noise_ratio=0.1):
        direction = direction + Vector2(((random.random() * 2 - 1)*noise_ratio)*direction.x, ((random.random() * 2 - 1)*noise_ratio)*direction.y)
        return direction



import os
import sys
import gym

from gobigger.server import Server
from gobigger.render import EnvRender


class GoBiggerEnv(gym.Env):

    def __init__(self, server_cfg=None):
        self.server_cfg = server_cfg
        self.server = Server(cfg=server_cfg)
        self.server.set_render(render=EnvRender(self.server.map_width, self.server.map_height))

    def step(self, actions):
        done = self.server.step(actions=actions)
        obs = self.server.obs()
        global_state, screen_data_players = obs
        total_size = [global_state['leaderboard'][str(i)] \
                        for i in range(len(global_state['leaderboard']))]
        assert len(self.last_total_size) == len(total_size)
        reward = [total_size[i] - self.last_total_size[i] for i in range(len(total_size))]
        self.last_total_size = total_size
        info = {}
        return obs, reward, done, info

    def reset(self):
        self.server.reset()
        obs = self.server.obs()
        global_state, screen_data_players = obs
        self.last_total_size = [global_state['leaderboard'][str(i)] \
                                for i in range(len(global_state['leaderboard']))]
        return obs

    def close(self):
        self.server.close()

    def seed(self, seed):
        self.server.seed(seed)



import logging
from easydict import EasyDict
from pygame.math import Vector2
import math

from gobigger.utils import format_vector, add_size, Border, deep_merge_dicts
from .base_ball import BaseBall


class SporeBall(BaseBall):
    """
    Overview:
        Spores spit out by the player ball
        - characteristic:
        * Can't move actively
        * can not eat
        * Can be eaten by CloneBall and ThornsBall
        * There is an initial velocity at birth, and it decays to 0 within a period of time
    """
    @staticmethod
    def default_config():
        cfg = BaseBall.default_config()
        cfg.update(dict(
            radius_min=3, # Minimum radius(Fixed)
            radius_max=3, # Maximum radius(Fixed)
            vel_init=50, # initial value if velocity(Fixed)
            vel_zero_time=0.1, # The time it takes for the speed to decay to zero(S)
            spore_radius_init=20, # initial radius(Fixed)
        ))
        return EasyDict(cfg)

    def __init__(self, name, position, border, size=1, vel=None, acc=None, direction=Vector2(0,0), **kwargs):
        # init other kwargs
        kwargs = EasyDict(kwargs)
        cfg = SporeBall.default_config()
        cfg = deep_merge_dicts(cfg, kwargs)
        super(SporeBall, self).__init__(name, position, border, size=size, vel=vel, acc=acc, **cfg)
        self.vel_init = cfg.vel_init
        self.vel_zero_time = cfg.vel_zero_time
        self.spore_radius_init = cfg.spore_radius_init
        # normal kwargs
        self.direction = direction.normalize()
        self.vel = self.vel_init * self.direction
        self.acc = - (self.vel_init / self.vel_zero_time) * self.direction
        self.move_time = 0
        # reset size
        if math.sqrt(self.size) != self.spore_radius_init:
            self.set_size(self.spore_radius_init**2)
        self.moving = True
        self.check_border()

    def move(self, direction=None, duration=0.05):
        assert direction is None
        assert duration > 0
        self.position = self.position + self.vel * duration
        self.move_time += duration
        if self.move_time < self.vel_zero_time:
            self.vel += self.acc * duration
        else:
            self.vel = Vector2(0, 0)
            self.acc = Vector2(0, 0)
            self.moving = False
        self.check_border()
        return True

    def eat(self, ball):
        logging.debug('SporeBall can not eat others')
        return



import math
import logging
from abc import ABC, abstractmethod
from easydict import EasyDict
from pygame.math import Vector2
from functools import total_ordering

from gobigger.utils import format_vector, Border, deep_merge_dicts


@total_ordering
class BaseBall(ABC):
    '''
    Overview:
        Base class of all balls
    '''
    @staticmethod
    def default_config():
        '''
        Overview:
            Default config
        '''
        cfg = dict(
            radius_min = 5, # Minimum radius
            radius_max = 10, # Maximum radius
        )
        return EasyDict(cfg)

    def __init__(self, name, position, border, size=None, vel=None, acc=None, color : int = 1, **kwargs):
        '''
        Parameters:
             border <border>: Circumscribed rectangle of the ball
             vel <Vector2> : the direction of the ball's speed 
             acc <Vector2> : the direction of the ball's acceleration
        '''
        self.name = name
        self.position = position
        self.border = border
        self.vel = Vector2(0, 0) if vel is None else vel 
        self.acc = Vector2(0, 0) if acc is None else acc 

        # init other kwargs
        kwargs = EasyDict(kwargs)
        cfg = BaseBall.default_config()
        cfg = deep_merge_dicts(cfg, kwargs)
        self.color=color
        self.radius_min = cfg.radius_min
        self.radius_max = cfg.radius_max
        if size is None:
            size = self.radius_min ** 2
        self.set_size(size)
        self.is_remove = False
        self.quad_node = None

    def set_size(self, size: float) -> None:
        """
        Overview:
            Set radius according to weight
        Parameters:
            size <float>: weight
        """
        self.size = size
        self.radius = math.sqrt(self.size)
        if self.radius > self.radius_max:
            self.radius = self.radius_max
        elif self.radius < self.radius_min:
            self.radius = self.radius_min
        self.size = self.radius * self.radius

    def move(self, direction, duration):
        """
        Overview:
            Realize the movement of the ball, pass in the direction and time parameters, and return the new position
        Parameters:
            direction <Vector2>: A point in the unit circle
            duration <float>: time
        Returns:
            position <Vector2>: position after moving 
        """
        raise NotImplementedError

    def eat(self, ball):
        """
        Overview:
            Describe the rules of eating and being eaten
        Parameters:
            ball <BaseBall>: Eaten ball
        """
        raise NotImplementedError

    def remove(self):
        """
        Overview:
            Things to do when being removed from the map
        """
        self.is_remove = True

    def check_border(self):
        """
        Overview:
            Check to see if the position of the ball exceeds the bounds of the map. 
            If it exceeds, the speed and acceleration in the corresponding direction will be zeroed, and the position will be edged
        """
        if self.position.x < self.border.minx + self.radius or self.position.x > self.border.maxx - self.radius:
            self.position.x = max(self.position.x, self.border.minx + self.radius)
            self.position.x = min(self.position.x, self.border.maxx - self.radius)
            self.vel.x = 0
            self.acc.x = 0
        if self.position.y < self.border.miny + self.radius or self.position.y > self.border.maxy - self.radius:
            self.position.y = max(self.position.y, self.border.miny + self.radius)
            self.position.y = min(self.position.y, self.border.maxy - self.radius)
            self.vel.y = 0
            self.acc.y = 0

    def get_dis(self, ball):
        '''
        Overview:
            Get the distance between the centers of the two balls
        Parameters:
            ball <BaseBall>: another ball
        '''
        return (self.position - ball.position).length()

    def judge_cover(self, ball):
        '''
        Overview:
            Determine whether the center of the two balls is covered
        Parameters:
            ball <BaseBall>: another ball
        Returns:
            is_covered <bool>: covered or not
        '''
        if ball.name == self.name:
            return False
        dis = self.get_dis(ball)
        if self.radius > dis or ball.radius > dis:
            return True
        else:
            return False

    def judge_in_rectangle(self, rectangle):
        '''
        Overview:
            Determine if the ball and rectangle intersect
        Parameters:
            rectangle <List>: left_top_x, left_top_y, right_bottom_x, right_bottom_y
        Returns:
            <bool> : intersect or not
        '''
        dx = rectangle[0] - self.position.x if rectangle[0] > self.position.x \
                else self.position.x - rectangle[2] if self.position.x > rectangle[2] else 0
        dy = rectangle[1] - self.position.y if rectangle[1] > self.position.y \
                else self.position.y - rectangle[3] if self.position.y > rectangle[3] else 0
        return dx**2 + dy**2 <= self.radius**2

    def __repr__(self) -> str:
        return 'position={}, size={:.3f}, radius={:.3f}, vel={}, acc={}'.format(self.position, self.size, self.radius, self.vel, self.acc)

    def __eq__(self, other):
        return self.size == other.size

    def __le__(self, other):
        return self.size < other.size

    def __gt__(self, other):
        return self.size > other.size



import math
from pygame.math import Vector2
import logging
import random
import cv2
import numpy as np


def format_vector(v, norm_max):
    '''
    Overview:
        The maximum value of the given vector's modulus
         example:
             1) The maximum speed limit is 5, given that the current speed is (6,8), it will return (3,4)
             2) Limit the maximum acceleration and return to the acceleration after the limit
    '''
    if v.length() < norm_max:
        # logging.debug('v={}, v.length()={}, norm_max={}'.format(v, v.length(), norm_max))
        return v
    else:
        return v.normalize() * norm_max


def add_size(size_old, size_add):
    '''
    Overview:
        Calculate the size of the big ball after eating the small ball
    '''
    return size_old + size_add


def save_screen_data_to_img(screen_data, img_path=None):
    '''
    Overview:
        Save the numpy screen data as the corresponding picture
    '''
    img = cv2.cvtColor(screen_data, cv2.COLOR_RGB2BGR)
    img = np.fliplr(img)
    img = np.rot90(img)
    if img_path is not None:
        cv2.imwrite(img_path, img)


class Border:
    '''
    Overview:
        used to specify a rectangular range
    '''
    def __init__(self, minx, miny, maxx, maxy, random_generator=None):
        self.minx = minx
        self.miny = miny
        self.maxx = maxx
        self.maxy = maxy
        self.width = self.maxx - self.minx
        self.height =self.maxy - self.miny
        if random_generator is not None:
            self._random = random_generator
        else:
            self._random = random.Random()

    def __repr__(self) -> str:
        return '[' + str(self.minx) + ',' + str(self.miny) + ',' + str(self.maxx) + ',' + str(self.maxy) + ']'

    def contains(self, position: Vector2) -> bool:
        '''
        Overview:
            To judge whether a position in this border.
        Parameters:
            position <Vector2>: the position to be judged.
        Returns:
            bool: True or False, whether the position in this border.
        '''
        return position.x > self.minx and position.x < self.maxx and position.y > self.miny and position.y < self.maxy

    def sample(self) -> Vector2:
        '''
        Overview:
            Randomly sample a position in the border.
        Returns:
            Vector2: the sampled position.
        '''
        x = self._random.uniform(self.minx, self.maxx)
        y = self._random.uniform(self.miny, self.maxy)
        return Vector2(x, y)

    def get_joint(self, border) :
        new_minx = max(self.minx, border.minx)
        new_maxx = min(self.maxx, border.maxx)
        new_miny = max(self.miny, border.miny)
        new_maxy = min(self.maxy, border.maxy)
        if new_minx > new_maxx or new_miny > new_maxy:
            return None
        return Border(new_minx, new_maxx, new_miny, new_maxy, self._random)


class QuadNode:
    def __init__(self, border, max_depth = 32, max_num = 64, parent = None) -> None:
        self.border = border
        self.max_depth = max_depth
        self.midx = (border.minx + border.maxx) / 2
        self.midy = (border.miny + border.maxy) / 2
        self.max_num = max_num
        self.children = None
        self.parent = parent
        self.items = []

    def get_quad(self, node):
        if node.position.x < self.midx:
            if node.position.y < self.midy : return 0
            else: return 1
        else: 
            if node.position.y < self.midy: return 2
            else: return 3

    def insert(self, node):
        if not self.children == None:
            self.children[self.get_quad(node)].insert(node)
        else:
            self.items.append(node)
            node.quad_node = self
            if len(self.items) > self.max_num and self.max_depth >= 1:
                b0 = Border(self.border.minx, self.border.miny, self.midx, self.midy)
                b1 = Border(self.border.minx, self.midy, self.midx, self.border.maxy)
                b2 = Border(self.midx, self.border.miny, self.border.maxx, self.midy)
                b3 = Border(self.midx, self.midy, self.border.maxx,  self.border.maxy)
                self.children = []
                self.children.append(QuadNode(b0, max_depth = self.max_depth - 1, max_num = self.max_num, parent=self))
                self.children.append(QuadNode(b1, max_depth = self.max_depth - 1, max_num = self.max_num, parent=self))
                self.children.append(QuadNode(b2, max_depth = self.max_depth - 1, max_num = self.max_num, parent=self))
                self.children.append(QuadNode(b3, max_depth = self.max_depth - 1, max_num = self.max_num, parent=self))
                for item in self.items:
                    self.children[self.get_quad(item)].insert(item)
                self.items.clear()
    
    def find(self, border):
        ans = self.items
        if not self.children == None:
            for child in self.children:
                tmpBorder = border.get_joint(child.border)
                if not tmpBorder == None:
                    ans = ans + child.find(tmpBorder)
        return ans

    def clear(self):
        if self.children == None: return
        max_num = self.max_num
        for child in self.children:
            if not child.children == None: return
            max_num = max_num - len(child.items)
        if max_num >= 0:
            for child in self.children:
                for item in child.items:
                    item.quad_node = self
                    self.items.append(item)
            self.children = None
            if not self.parent == None:
                self.parent.clear()
            
    def remove(self, node):
        for i, item in enumerate(self.items):
            if item.name == node.name:
                del self.items[i]
                break
        node.quad_node = None
        if not self.parent == None:
            self.parent.clear()



import logging
import pytest
import uuid
from pygame.math import Vector2
import pygame
import numpy as np
import cv2
import time

from .base_render import BaseRender
from gobigger.utils import FOOD_COLOR, THORNS_COLOR, SPORE_COLOR, PLAYER_COLORS, BACKGROUND, BLACK, WHITE
from gobigger.utils import FOOD_COLOR_GRAYSCALE, THORNS_COLOR_GRAYSCALE, SPORE_COLOR_GRAYSCALE, PLAYER_COLORS_GRAYSCALE, BACKGROUND_GRAYSCALE
from gobigger.utils import to_aliased_circle


class EnvRender(BaseRender):
    '''
    Overview:
        No need to use a new window, giving a global view and the view that each player can see
    '''
    def __init__(self, width, height, padding=(0,0), cell_size=10, 
                 scale_up_ratio=1.5, vision_x_min=100, vision_y_min=100, only_render=True):
        super(EnvRender, self).__init__(width, height, padding=padding, 
                                        cell_size=cell_size, only_render=only_render)
        self.scale_up_ratio = scale_up_ratio
        self.vision_x_min = vision_x_min
        self.vision_y_min = vision_y_min

    def set_obs_settings(self, obs_settings):
        self.with_spatial = obs_settings.get('with_spatial', True)
        self.with_speed = obs_settings.get('with_speed', False)
        self.with_all_vision = obs_settings.get('with_all_vision', False)
        self.cheat = obs_settings.get('cheat', False)

    def fill_all(self, screen, food_balls, thorns_balls, spore_balls, players):
        font = pygame.font.SysFont('Menlo', 12, True)
        # render all balls
        for ball in food_balls:
            pygame.draw.circle(screen, FOOD_COLOR_GRAYSCALE, ball.position, ball.radius)
        for ball in thorns_balls:
            pygame.draw.polygon(screen, THORNS_COLOR_GRAYSCALE, to_aliased_circle(ball.position, ball.radius))
        for ball in spore_balls:
            pygame.draw.circle(screen, SPORE_COLOR_GRAYSCALE, ball.position, ball.radius)
        for index, player in enumerate(players):
            for ball in player.get_balls():
                pygame.draw.circle(screen, PLAYER_COLORS_GRAYSCALE[int(ball.owner)], ball.position, ball.radius)
        screen_data = pygame.surfarray.array2d(screen)
        return screen_data

    def get_clip_screen(self, screen_data, rectangle):
        if len(screen_data.shape) == 3:
            screen_data_clip = screen_data[rectangle[0]:rectangle[2], 
                                           rectangle[1]:rectangle[3], :]
        elif len(screen_data.shape) == 2:
            screen_data_clip = screen_data[rectangle[0]:rectangle[2], 
                                           rectangle[1]:rectangle[3]]
        else:
            raise NotImplementedError
        return screen_data_clip

    def get_rectangle_by_player(self, player):
        '''
        Multiples of the circumscribed matrix of the centroid
        '''
        centroid = player.cal_centroid()
        xs_max = 0
        ys_max = 0
        for ball in player.get_balls():
            direction_center = centroid - ball.position
            if abs(direction_center.x) + ball.radius > xs_max:
                xs_max = abs(direction_center.x) + ball.radius
            if abs(direction_center.y) + ball.radius > ys_max:
                ys_max = abs(direction_center.y) + ball.radius
        xs_max = max(xs_max, self.vision_x_min)
        ys_max = max(ys_max, self.vision_y_min)
        scale_up_len =  max(xs_max, ys_max)
        left_top_x = min(max(int(centroid.x - scale_up_len * self.scale_up_ratio), 0), 
                         max(int(self.width_full - scale_up_len * self.scale_up_ratio * 2), 0))
        left_top_y = min(max(int(centroid.y - scale_up_len * self.scale_up_ratio), 0),
                         max(int(self.height_full - scale_up_len * self.scale_up_ratio * 2), 0))
        right_bottom_x = min(int(left_top_x + scale_up_len * self.scale_up_ratio * 2), self.width_full)
        right_bottom_y = min(int(left_top_y + scale_up_len * self.scale_up_ratio * 2), self.height_full)
        rectangle = (left_top_x, left_top_y, right_bottom_x, right_bottom_y)
        return rectangle

    def get_overlap(self, rectangle, food_balls, thorns_balls, spore_balls, players):
        ret = {}
        food_count = 0
        thorns_count = 0
        spore_count = 0
        clone_count = 0

        assert len(players) > 0, 'len(players) = {} can not be 0'.format(len(players))

        food = len(food_balls) * [3 * [None]]     # without speed
        thorns = len(thorns_balls) * [3 * [None]] # without speed 
        spore = len(spore_balls) * [3 * [None]]   # without speed 
        clone = len(players) * players[0].ball_settings.part_num_max * [5 * [None]]   # without speed 
        # food overlap
        if len(food_balls) > 0:
            food_radius = food_balls[0].radius
            fr0 = rectangle[0] - food_radius
            fr1 = rectangle[1] - food_radius
            fr2 = rectangle[2] + food_radius
            fr3 = rectangle[3] + food_radius
        for ball in food_balls:
            x = ball.position.x
            y = ball.position.y
            if x < fr0 or x > fr2 or y < fr1 or y > fr3:
                continue
            else:
                food[food_count] = [x, y, ball.radius]
                food_count += 1
        food = food[:food_count]
        ret['food'] = food

        # thorns overlap
        for ball in thorns_balls:
            if ball.judge_in_rectangle(rectangle):
                thorns[thorns_count] = [ball.position.x, ball.position.y, ball.radius]
                thorns_count += 1
        thorns = thorns[:thorns_count]
        ret['thorns'] = thorns
        # spore overlap
        for ball in spore_balls:
            if ball.judge_in_rectangle(rectangle):
                spore[spore_count] = [ball.position.x, ball.position.y, ball.radius]
                spore_count += 1
        spore = spore[:spore_count]
        ret['spore'] = spore
        # clone overlap
        for player in players:
            for ball in player.get_balls():
                if ball.judge_in_rectangle(rectangle):
                    clone[clone_count] = [ball.position.x, ball.position.y, ball.radius, player.name, player.team_name]
                    clone_count += 1
        clone = clone[:clone_count]
        ret['clone'] = clone

        return ret
    
    def get_overlap_wo_rectangle(self, food_balls, thorns_balls, spore_balls, players):
        ret = {'food': [], 'thorns': [], 'spore': [], 'clone': []}
        for ball in food_balls:
            ret['food'].append([ball.position.x, ball.position.y, ball.radius])
        for ball in thorns_balls:
            ret['thorns'].append([ball.position.x, ball.position.y, ball.radius])
        for ball in spore_balls:
            ret['spore'].append([ball.position.x, ball.position.y, ball.radius])
        for player in players:
            for ball in player.get_balls():
                ret['clone'].append([ball.position.x, ball.position.y, ball.radius, int(player.name), int(player.team_name)])
        return ret

    def get_overlap_wo_rectangle_with_speed(self, food_balls, thorns_balls, spore_balls, players):
        ret = {'food': [], 'thorns': [], 'spore': [], 'clone': []}
        for ball in food_balls:
            ret['food'].append([ball.position.x, ball.position.y, ball.radius])
        for ball in thorns_balls:
            ret['thorns'].append([ball.position.x, ball.position.y, ball.radius, ball.vel.x, ball.vel.y])
        for ball in spore_balls:
            ret['spore'].append([ball.position.x, ball.position.y, ball.radius, ball.vel.x, ball.vel.y])
        for player in players:
            for ball in player.get_balls():
                ret['clone'].append([ball.position.x, ball.position.y, ball.radius, ball.vel.x+ball.vel_last.x, ball.vel.y+ball.vel_last.y, int(player.name), int(player.team_name)])
        return ret

    def get_overlap_with_speed(self, rectangle, food_balls, thorns_balls, spore_balls, players):
        ret = {}
        food_count = 0
        thorns_count = 0
        spore_count = 0
        clone_count = 0

        assert len(players) > 0, 'len(players) = {} can not be 0'.format(len(players))

        food = len(food_balls) * [3 * [None]]     # without speed, as food can not move
        thorns = len(thorns_balls) * [5 * [None]] # with speed 
        spore = len(spore_balls) * [5 * [None]]   # with speed 
        clone = len(players) * players[0].ball_settings.part_num_max * [7 * [None]]   # with speed 
        # food overlap
        if len(food_balls) > 0:
            food_radius = food_balls[0].radius
            fr0 = rectangle[0] - food_radius
            fr1 = rectangle[1] - food_radius
            fr2 = rectangle[2] + food_radius
            fr3 = rectangle[3] + food_radius
        for ball in food_balls:
            x = ball.position.x
            y = ball.position.y
            if x < fr0 or x > fr2 or y < fr1 or y > fr3:
                continue
            else:
                food[food_count] = [x, y, ball.radius]
                food_count += 1
        food = food[:food_count]
        ret['food'] = food
        # thorns overlap
        for ball in thorns_balls:
            if ball.judge_in_rectangle(rectangle):
                thorns[thorns_count] = [ball.position.x, ball.position.y, ball.radius, ball.vel.x, ball.vel.y]
                thorns_count += 1
        thorns = thorns[:thorns_count]
        ret['thorns'] = thorns
        # spore overlap
        for ball in spore_balls:
            if ball.judge_in_rectangle(rectangle):
                spore[spore_count] = [ball.position.x, ball.position.y, ball.radius, ball.vel.x, ball.vel.y]
                spore_count += 1
        spore = spore[:spore_count]
        ret['spore'] = spore
        # clone overlap
        for player in players:
            for ball in player.get_balls():
                if ball.judge_in_rectangle(rectangle):
                    clone[clone_count] = [ball.position.x, ball.position.y, ball.radius, ball.vel.x+ball.vel_last.x, ball.vel.y+ball.vel_last.y, player.name, player.team_name]
                    clone_count += 1
        clone = clone[:clone_count]
        ret['clone'] = clone

        return ret

    def update_all(self, food_balls, thorns_balls, spore_balls, players):
        screen_data_all = None
        feature_layers = None
        overlap = None
        rectangle = None
        if self.with_spatial:
            screen_all = pygame.Surface((self.width, self.height), depth=8)
            screen_all.fill(BACKGROUND_GRAYSCALE)
            screen_data_all = self.fill_all(screen_all, food_balls, thorns_balls, spore_balls, players)
        screen_data_players = {}

        if self.cheat: # all map and all players' vision
            for player in players:
                rectangle = self.get_rectangle_by_player(player)
                if self.with_spatial:
                    screen_data_player = self.get_clip_screen(screen_data_all, rectangle=rectangle)
                    screen_data_player = np.fliplr(screen_data_player)
                    screen_data_player = np.rot90(screen_data_player)
                    feature_layers = self.transfer_rgb_to_features(screen_data_player, player_num=len(players))
                if self.with_speed:
                    overlap = self.get_overlap_with_speed(rectangle, food_balls, thorns_balls, spore_balls, players)
                else:
                    overlap = self.get_overlap(rectangle, food_balls, thorns_balls, spore_balls, players)
                screen_data_players[player.name] = {
                    'feature_layers': feature_layers,
                    'rectangle': rectangle,
                    'overlap': overlap,
                    'team_name': player.team_name,
                }

            # all map info
            if self.with_spatial:
                screen_data_player = np.fliplr(screen_data_all)
                screen_data_player = np.rot90(screen_data_player)
                feature_layers = self.transfer_rgb_to_features(screen_data_player, player_num=len(players))
            if self.with_speed:
                overlap = self.get_overlap_wo_rectangle_with_speed(food_balls, thorns_balls, spore_balls, players)
            else:
                overlap = self.get_overlap_wo_rectangle(food_balls, thorns_balls, spore_balls, players)
            screen_data_players['all'] = {
                'feature_layers': feature_layers,
                'rectangle': [0, 0, self.width, self.height],
                'overlap': overlap,
                'team_name': '',
            }

        elif self.with_all_vision:
            for player in players:
                if player.name == '0':
                    if self.with_spatial:
                        screen_data_player = np.fliplr(screen_data_all)
                        screen_data_player = np.rot90(screen_data_player)
                        feature_layers = self.transfer_rgb_to_features(screen_data_player, player_num=len(players))
                    if self.with_speed:
                        overlap = self.get_overlap_wo_rectangle_with_speed(food_balls, thorns_balls, spore_balls, players)
                    else:
                        overlap = self.get_overlap_wo_rectangle(food_balls, thorns_balls, spore_balls, players)
                    screen_data_players[player.name] = {
                        'feature_layers': feature_layers,
                        'rectangle': [0,0,self.width,self.height],
                        'overlap': overlap,
                        'team_name': player.team_name,
                    }
                else:
                    screen_data_players[player.name] = {
                        'feature_layers': None,
                        'rectangle': None,
                        'overlap': None,
                        'team_name': player.team_name,
                    }
        else:
            for player in players:
                rectangle = self.get_rectangle_by_player(player)
                if self.with_spatial:
                    screen_data_player = self.get_clip_screen(screen_data_all, rectangle=rectangle)
                    screen_data_player = np.fliplr(screen_data_player)
                    screen_data_player = np.rot90(screen_data_player)
                    feature_layers = self.transfer_rgb_to_features(screen_data_player, player_num=len(players))
                if self.with_speed:
                    overlap = self.get_overlap_with_speed(rectangle, food_balls, thorns_balls, spore_balls, players)
                else:
                    overlap = self.get_overlap(rectangle, food_balls, thorns_balls, spore_balls, players)
                screen_data_players[player.name] = {
                    'feature_layers': feature_layers,
                    'rectangle': rectangle,
                    'overlap': overlap,
                    'team_name': player.team_name,
                }
        return screen_data_all, screen_data_players

    def render_all_balls_colorful(self, screen, food_balls, thorns_balls, spore_balls, players, player_num_per_team):
        # render all balls
        for ball in food_balls:
            pygame.draw.circle(screen, FOOD_COLOR, ball.position, ball.radius)
        for ball in thorns_balls:
            pygame.draw.polygon(screen, THORNS_COLOR, to_aliased_circle(ball.position, ball.radius))
        for ball in spore_balls:
            pygame.draw.circle(screen, SPORE_COLOR, ball.position, ball.radius)
        player_name_size = {}
        for index, player in enumerate(players):
            for ball in player.get_balls():
                pygame.draw.circle(screen, PLAYER_COLORS[int(ball.team_name)][0], ball.position, ball.radius)
                font_size = int(ball.radius/1.6)
                font = pygame.font.SysFont('arial', max(font_size, 8), True)
                txt = font.render('{}'.format(chr(int(ball.owner)%player_num_per_team+65)), True, WHITE)
                txt_rect = txt.get_rect(center=(ball.position.x, ball.position.y))
                screen.blit(txt, txt_rect)
            player_name_size[player.name] = player.get_total_size()
        return screen, player_name_size

    def render_leaderboard_colorful(self, screen, team_name_size, player_name_size, player_num_per_team):
        team_name_size = sorted(team_name_size.items(), key=lambda d: d[1], reverse=True)
        start = 10
        for index, (team_name, size) in enumerate(team_name_size):
            start += 20
            font = pygame.font.SysFont('arial', 16, True)
            fps_txt = font.render('{} : {:.3f}'.format(team_name, size), True, PLAYER_COLORS[int(team_name)][0])
            screen.blit(fps_txt, (self.width+20, start))
            start += 20
            font = pygame.font.SysFont('arial', 14, True)
            for j in range(player_num_per_team):
                player_name = str(int(team_name)*player_num_per_team+j)
                player_size = player_name_size[player_name]
                fps_txt = font.render('  {} : {:.3f}'.format(chr(int(player_name)%player_num_per_team+65), player_size), True, PLAYER_COLORS[int(team_name)][0])
                screen.blit(fps_txt, (self.width+20, start))
                start += 20
        return screen

    def get_tick_all_colorful(self, food_balls, thorns_balls, spore_balls, players, partial_size=300, player_num_per_team=3, 
                              bar_width=150, team_name_size=None):
        screen_all = pygame.Surface((self.width+bar_width, self.height))
        screen_all.fill(BACKGROUND)
        pygame.draw.line(screen_all, BLACK, (self.width+1, 0), (self.width+1, self.height), width=3)
        screen_all, player_name_size = self.render_all_balls_colorful(screen_all, food_balls, thorns_balls, spore_balls, players, 
                                                                      player_num_per_team=player_num_per_team)
        screen_all = self.render_leaderboard_colorful(screen_all, team_name_size, player_name_size, player_num_per_team=player_num_per_team)
        screen_data_all = pygame.surfarray.array3d(screen_all)
        screen_data_players = {}
        for player in players:
            rectangle = self.get_rectangle_by_player(player)
            screen_data_player = self.get_clip_screen(screen_data_all, rectangle=rectangle)
            screen_data_player = cv2.resize(np.rot90(np.fliplr(cv2.cvtColor(screen_data_player, cv2.COLOR_RGB2BGR))), (partial_size, partial_size))
            screen_data_players[player.name] = screen_data_player
        screen_data_all = np.rot90(np.fliplr(cv2.cvtColor(screen_data_all, cv2.COLOR_RGB2BGR)))
        return screen_data_all, screen_data_players

    def transfer_rgb_to_features(self, rgb, player_num=12):
        '''
        Overview:
            If player_num == 12, then the features list will contain 15 elements(12 player + food + spore + thorns)
        '''
        features = []
        assert len(rgb.shape) == 2
        h, w = rgb.shape
        for i in range(player_num):
            features.append((rgb==PLAYER_COLORS_GRAYSCALE[i]).astype(int))
        features.append((rgb==FOOD_COLOR_GRAYSCALE).astype(int))
        features.append((rgb==SPORE_COLOR_GRAYSCALE).astype(int))
        features.append((rgb==THORNS_COLOR_GRAYSCALE).astype(int))
        return features

    def show(self):
        raise NotImplementedError

    def close(self):
        pygame.quit()



import os
import logging


class BaseAgent:
    '''
    Overview:
        The base class of all agents
    '''
    def __init__(self):
        pass

    def step(self, obs):
        raise NotImplementedError
        



import copy
from typing import List, Optional


def deep_merge_dicts(original: dict, new_dict: dict) -> dict:
    """
    Overview:
        Merge two dicts by calling ``deep_update``
    Arguments:
        - original (:obj:`dict`): Dict 1.
        - new_dict (:obj:`dict`): Dict 2.
    Returns:
        - merged_dict (:obj:`dict`): A new dict that is d1 and d2 deeply merged.
    """
    original = original or {}
    new_dict = new_dict or {}
    merged = copy.deepcopy(original)
    if new_dict:  # if new_dict is neither empty dict nor None
        deep_update(merged, new_dict, True, [])
    return merged


def deep_update(
    original: dict,
    new_dict: dict,
    new_keys_allowed: bool = False,
    whitelist: Optional[List[str]] = None,
    override_all_if_type_changes: Optional[List[str]] = None
):
    """
    Overview:
        Update original dict with values from new_dict recursively.
    Arguments:
        - original (:obj:`dict`): Dictionary with default values.
        - new_dict (:obj:`dict`): Dictionary with values to be updated
        - new_keys_allowed (:obj:`bool`): Whether new keys are allowed.
        - whitelist (:obj:`Optional[List[str]]`):
            List of keys that correspond to dict
            values where new subkeys can be introduced. This is only at the top
            level.
        - override_all_if_type_changes(:obj:`Optional[List[str]]`):
            List of top level
            keys with value=dict, for which we always simply override the
            entire value (:obj:`dict`), if the "type" key in that value dict changes.

    .. note::

        If new key is introduced in new_dict, then if new_keys_allowed is not
        True, an error will be thrown. Further, for sub-dicts, if the key is
        in the whitelist, then new subkeys can be introduced.
    """
    whitelist = whitelist or []
    override_all_if_type_changes = override_all_if_type_changes or []

    for k, value in new_dict.items():
        if k not in original and not new_keys_allowed:
            raise RuntimeError("Unknown config parameter `{}`. Base config have: {}.".format(k, original.keys()))

        # Both original value and new one are dicts.
        if isinstance(original.get(k), dict) and isinstance(value, dict):
            # Check old type vs old one. If different, override entire value.
            if k in override_all_if_type_changes and \
                    "type" in value and "type" in original[k] and \
                    value["type"] != original[k]["type"]:
                original[k] = value
            # Whitelisted key -> ok to add new subkeys.
            elif k in whitelist:
                deep_update(original[k], value, True)
            # Non-whitelisted key.
            else:
                deep_update(original[k], value, new_keys_allowed)
        # Original value not a dict OR new value not a dict:
        # Override entire value.
        else:
            original[k] = value
    return original



from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

from setuptools import setup


setup(
    name='gobigger',
    version='0.1.5',
    description='Go-Bigger: Multi-Agent Decision Intelligence Environment',
    author='OpenDILab',
    license='Apache License, Version 2.0',
    keywords='Go-Bigger DI',
    packages=[
        'gobigger',
        'gobigger.agents',
        'gobigger.balls',
        'gobigger.server',
        'gobigger.utils',
        'gobigger.managers',
        'gobigger.players',
        'gobigger.render',
        'gobigger.envs',
        'gobigger.bin',
        'gobigger.hyper',
    ],
    install_requires=[
        'easydict',
        'gym>=0.15.3',  # pypy incompatible
        'pygame>=2.0.0',
        'pytest>=5.0.0',
        'opencv-python',
        'numpy>=1.10, <= 1.19',
    ]
)



import random
import math
import logging
import uuid
from abc import ABC, abstractmethod
from easydict import EasyDict
from pygame.math import Vector2

from .base_manager import BaseManager
from gobigger.utils import format_vector, Border
from gobigger.balls import FoodBall, ThornsBall, CloneBall, SporeBall
from gobigger.players import HumanPlayer


class PlayerManager(BaseManager):

    def __init__(self, cfg, border, team_num, player_num_per_team, spore_manager_settings, random_generator=None):
        super(PlayerManager, self).__init__(cfg, border)
        self.players = {}
        self.team_num = team_num
        self.player_num_per_team = player_num_per_team
        self.player_num = self.team_num * self.player_num_per_team
        self.spore_manager_settings = spore_manager_settings
        self.spore_settings = self.spore_manager_settings.ball_settings
        if random_generator is not None:
            self._random = random_generator
        else:
            self._random = random.Random()

    def init_balls(self, custom_init=None):
        if custom_init is None:
            for i in range(self.team_num):
                team_name = str(i)
                for j in range(self.player_num_per_team):
                    player_name = str(i * self.player_num_per_team + j)
                    player = HumanPlayer(cfg=self.cfg.ball_settings, team_name=team_name, name=player_name, border=self.border, spore_settings=self.spore_settings)
                    player.respawn(position=self.border.sample())
                    self.players[player_name] = player
        else:
            init_dict = {}
            for i in range(self.team_num):
                team_name = str(i)
                init_dict[team_name] = {}
                for j in range(self.player_num_per_team):
                    player_name = str(i * self.player_num_per_team + j)
                    player = HumanPlayer(cfg=self.cfg.ball_settings, team_name=team_name, name=player_name, border=self.border, spore_settings=self.spore_settings)
                    self.players[player_name] = player
                    init_dict[team_name][player_name] = False
            for ball_cfg in custom_init:
                # [position.x, position.y, radius, player_name, team_name, 
                #  vel.x, vel.y, acc.x, acc.y, vel_last.x,
                #  vel_last.y, acc_last.x, acc_last.y, direction.x, direction.y,
                #  last_given_acc.x, last_given_acc.y, age, cooling_last, stop_flag,
                #  stop_time, acc_stop.x, acc_stop.y]
                position = Vector2(*ball_cfg[0:2])
                radius = ball_cfg[2]
                player_name = ball_cfg[3]
                team_name = ball_cfg[4]
                ball = CloneBall(team_name=team_name, name=uuid.uuid1(), position=position, border=self.border, 
                                 size=radius**2, vel=Vector2(0,0), acc=Vector2(0,0),
                                 vel_last=Vector2(0,0), acc_last=Vector2(0,0), last_given_acc=Vector2(0,0),
                                 stop_flag=True, owner=player_name, spore_settings=self.spore_settings, **self.cfg.ball_settings)
                if len(ball_cfg) > 5:
                    ball.vel = Vector2(*ball_cfg[5:7])
                    ball.acc = Vector2(*ball_cfg[7:9])
                    ball.vel_last = Vector2(*ball_cfg[9:11])
                    ball.acc_last = Vector2(*ball_cfg[11:13])
                    ball.direction = Vector2(*ball_cfg[13:15])
                    ball.last_given_acc = Vector2(*ball_cfg[15:17])
                    ball.age = ball_cfg[17]
                    ball.cooling_last = ball_cfg[18]
                    ball.stop_flag = ball_cfg[19]
                    ball.stop_time = ball_cfg[20]
                    ball.acc_stop = Vector2(*ball_cfg[21:23])
                self.players[player_name].add_balls(ball)
                init_dict[team_name][player_name] = True
            for team_name, team in init_dict.items():
                for player_name, player_init_flag in team.items():
                    if not player_init_flag:
                        self.players[player_name].respawn(position=self.border.sample())
            
    def get_balls(self):
        balls = []
        for player_name, player in self.players.items():
            balls.extend(player.get_balls())
        return balls

    def get_players(self):
        return list(self.players.values())

    def get_player_by_name(self, name):
        assert name in self.players
        return self.players[name]
    
    def add_balls(self, balls):
        if isinstance(balls, list):
            for ball in balls:
                self.players[ball.owner].add_balls(ball)
        elif isinstance(balls, CloneBall):
            self.players[balls.owner].add_balls(balls)
        return True

    def remove_balls(self, balls):
        if isinstance(balls, list):
            for ball in balls:
                self.players[ball.owner].remove_balls(ball)
        elif isinstance(balls, CloneBall):
            self.players[balls.owner].remove_balls(balls)

    def step(self):
        for player_name, player in self.players.items():
            if player.get_clone_num() == 0:
                player.respawn(position=self.border.sample())

    def adjust(self):
        '''
        Overview:
            Adjust all balls in all players
        '''
        for player in self.get_players():
            player.adjust()
        return True

    def get_clone_num(self, ball):
        return self.players[ball.owner].get_clone_num()

    def get_player_names(self):
        '''
        Overview:
            get all names of players
        '''
        return [player.name for player in self.get_players()]

    def get_team_names(self):
        '''
        Overview:
            get all names of players by teams with team names
        '''
        ret = {}
        for player in self.get_players():
            if player.team_name not in ret:
                ret[player.team_name] = []
            ret[player.team_name].append(player.name)
        return ret

    def get_player_names_with_team(self):
        '''
        Overview:
            get all names of players by teams
        '''
        ret = {}
        for player in self.get_players():
            if player.team_name not in ret:
                ret[player.team_name] = []
            ret[player.team_name].append(player.name)
        return list(ret.values())

    def get_teams_size(self):
        team_name_size = {}
        for player in self.get_players():
            if player.team_name not in team_name_size:
                team_name_size[player.team_name] = player.get_total_size()
            else:
                team_name_size[player.team_name] += player.get_total_size()
        return team_name_size

    def reset(self):
        '''
        Overview:
            reset manager
        '''
        self.refresh_time_count = 0
        self.players = {}
        return True



import numpy as np
import logging

from .structures import Border, QuadNode


class BaseCollisionDetection:

    def __init__(self, border: Border) -> None:
        self.border = border

    def solve(self, query_list: list, gallery_list: list):
        raise NotImplementedError


class ExhaustiveCollisionDetection(BaseCollisionDetection):
    '''
    Overview:
        Exhaustive Algorithm
    '''
    def __init__(self, border: Border) -> None:
        super(ExhaustiveCollisionDetection, self).__init__(border=border)

    def solve(self, query_list: list, gallery_list: list):
        '''
        Overview:
            For the balls in the query, enumerate each ball in the gallery to determine whether there is a collision
        Parameters:
            query_list <List[BaseBall]>: List of balls that need to be queried for collision
            gallery_list <List[BaseBall]>: List of all balls
        Returns:
            results <Dict[int: List[BaseBall]> return value
                int value denotes:
                    the subscript in query_list
                string value denotes:
                    List of balls that collided with the query corresponding to the subscript
        '''
        results = {}
        for i, q in enumerate(query_list):
            results[i] = []
            for j, g in enumerate(gallery_list):
                if q.judge_cover(g): 
                    results[i].append(g)
        return results


class PrecisionCollisionDetection(BaseCollisionDetection) :
    '''
    Overview:
        Precision Approximation Algorithm
        Divide the map into several rows according to the accuracy that has been set, dynamically maintain the row information in each frame, and search by row
    '''
    def __init__(self, border: Border, precision : int = 50) -> None:
        '''
        Parameter:
            precision <int>: the precision of dividing rows
        '''
        super(PrecisionCollisionDetection, self).__init__(border = border)
        self.precision = precision

    def get_row(self, x) -> int:
        '''
        Overview:
            Get the row coordinates of the ball
        Parameter:
            node <BaseBall>: The ball need to get its row coordinates
        '''
        return int((x - self.border.minx) / self.border.height * self.precision)


    def solve(self, query_list: list, gallery_list: list):
        '''
        Overview:
            First, you need to sort the balls in each row according to the ordinate. 
            For the balls in query_list, first abstract the boundary of the ball into 
            a rectangle, then traverse each row in the rectangle, and find the first 
            ball covered by the query through dichotomy in each row, and then Enumerate 
            the balls in sequence until the ordinate exceeds the boundary of the query 
            rectangle.
        Parameters:
            query_list <List[BaseBall]>: List of balls that need to be queried for collision
            gallery_list <List[BaseBall]>: List of all balls
        Returns:
            results <Dict[int: List[BaseBall]> return value
                int value denotes:
                    the subscript in query_list
                string value denotes:
                    List of balls that collided with the query corresponding to the subscript
        '''

        vec = {}
        for id, node in enumerate(gallery_list): 
            row_id = self.get_row(node.position.x)
            if  row_id not in vec:
                vec[row_id] = []
            vec[row_id].append((id, node.position.y))
        for val in vec.values():
            val.sort(key = lambda x: x[1])
        results = {}
        for id, query in enumerate(query_list):
            results[id] = []
            left = query.position.y - query.radius
            right = query.position.y + query.radius
            top = self.get_row(query.position.x - query.radius)
            bottom = self.get_row(query.position.x + query.radius)
            for i in range(top, bottom + 1):
                if i not in vec: continue
                l = len(vec[i])
                start_pos = 0
                for j in range(15, -1, -1):
                    if start_pos+(2**j) < l and vec[i][start_pos+(2**j)][1] < left:
                        start_pos += 2**j
                for j in range(start_pos, l):
                    if vec[i][j][1] > right: break
                    if query.judge_cover(gallery_list[vec[i][j][0]]):
                        results[id].append(gallery_list[vec[i][j][0]])
        return results


class RebuildQuadTreeCollisionDetection(BaseCollisionDetection) :
    '''
        Overview:
            Build a quadtree on a two-dimensional plane in every frame, and query collisions in the quadtree

    '''
    def __init__(self, border: Border, node_capacity = 64, tree_depth = 32) -> None:
        '''
        Parameter:
            node_capacity <int>: The capacity of each point in the quadtree
            tree_depth <int>: The max depth of the quadtree
        '''
        super(RebuildQuadTreeCollisionDetection, self).__init__(border = border)
        self.node_capacity = node_capacity
        self.tree_depth = tree_depth
        self.border = border
    
    def solve(self, query_list: list, gallery_list: list):
        '''
        Overview:
           Construct a quadtree from scratch based on gallery_list and complete the query
        Parameters:
            query_list <List[BaseBall]>: List of balls that need to be queried for collision
            gallery_list <List[BaseBall]>: List of all balls
        Returns:
            results <Dict[int: List[BaseBall]> return value
                int value denotes:
                    the subscript in query_list
                string value denotes:
                    List of balls that collided with the query corresponding to the subscript
        '''
        quadTree = QuadNode(border=self.border, max_depth = self.tree_depth, max_num = self.node_capacity)
        for node in gallery_list:
            quadTree.insert(node)
        results = {}
        for i, query in enumerate(query_list):
            results[i] = []
            quadTree_results = quadTree.find(Border(max(query.position.x - query.radius, self.border.minx), 
                                                    max(query.position.y - query.radius, self.border.miny), 
                                                    min(query.position.x + query.radius, self.border.maxx),
                                                    min(query.position.y + query.radius, self.border.maxy)))
            for result in quadTree_results:
                if query.judge_cover(result):
                    results[i].append(result)
        return results


class RemoveQuadTreeCollisionDetection(BaseCollisionDetection) :
    '''
        Overview:
            Add delete operations for the quadtree, and dynamically maintain a quadtree

    '''
    def __init__(self, border: Border, node_capacity = 64, tree_depth = 32) -> None:
        '''
        Parameter:
            node_capacity <int>: The capacity of each point in the quadtree
            tree_depth <int>: The max depth of the quadtree
        '''
        super(RemoveQuadTreeCollisionDetection, self).__init__(border = border)
        self.node_capacity = node_capacity
        self.tree_depth = tree_depth
        self.border = border
        self.quadTree = QuadNode(border = border, max_depth = tree_depth, max_num = node_capacity, parent=None)
    
    def solve(self, query_list: list, changed_node_list: list):
        '''
        Overview:
           Update the points in the quadtree according to the changed_node_list and complete the query
        Parameters:
            query_list <List[BaseBall]>: List of balls that need to be queried for collision
            gallery_list <List[BaseBall]>: List of all balls
        Returns:
            results <Dict[int: List[BaseBall]> return value
                int value denotes:
                    the subscript in query_list
                string value denotes:
                    List of balls that collided with the query corresponding to the subscript
        '''
        for node in changed_node_list:
            if not node.quad_node == None : node.quad_node.remove(node)
            if not node.is_remove: self.quadTree.insert(node)
        results = {}
        for i, query in enumerate(query_list):
            results[i] = []
            quadTree_results = self.quadTree.find(Border(max(query.position.x - query.radius, self.border.minx), 
                                                    max(query.position.y - query.radius, self.border.miny), 
                                                    min(query.position.x + query.radius, self.border.maxx),
                                                    min(query.position.y + query.radius, self.border.maxy)))
            for result in quadTree_results:
                if query.judge_cover(result):
                    results[i].append(result)
        return results

def create_collision_detection(cd_type, **cd_kwargs):
    if cd_type == 'exhaustive':
        return ExhaustiveCollisionDetection(**cd_kwargs)
    if cd_type == 'precision':
        return PrecisionCollisionDetection(**cd_kwargs)
    if cd_type == 'rebuild_quadtree':
        return RebuildQuadTreeCollisionDetection(**cd_kwargs)
    if cd_type == 'remove_quadtree':
        return RemoveQuadTreeCollisionDetection(**cd_kwargs)
    else:
        raise NotImplementedError




import math
import logging
import uuid
import copy
from easydict import EasyDict
from pygame.math import Vector2

from gobigger.utils import format_vector, add_size, Border, deep_merge_dicts
from .base_ball import BaseBall
from .food_ball import FoodBall
from .thorns_ball import ThornsBall
from .spore_ball import SporeBall


class CloneBall(BaseBall):
    """
    Overview:
        One of the balls that a single player can control
        - characteristic:
        * Can move
        * Can eat any other ball smaller than itself
        * Under the control of the player, the movement can be stopped immediately and contracted towards the center of mass of the player
        * Skill 1: Split each unit into two equally
        * Skill 2: Spit spores forward
        * There is a percentage of weight attenuation, and the radius will shrink as the weight attenuates
    """
    @staticmethod
    def default_config():

        cfg = BaseBall.default_config()
        cfg.update(dict(
            acc_max=50, # Maximum acceleration
            vel_max=25, # Maximum velocity
            radius_min=2, # Minimum radius
            radius_max=100, # Maximum radius
            radius_init=2, # The initial radius of the player's ball
            part_num_max=16, # Maximum number of avatars
            on_thorns_part_num=10, # Maximum number of splits when encountering thorns
            on_thorns_part_radius_max=5, # Maximum radius of split part when encountering thorns
            split_radius_min=30, # The lower limit of the radius of the splittable ball
            eject_radius_min=30, # The lower limit of the radius of the ball that can spores
            recombine_age=30, # Time for the split ball to rejoin (s)
            split_vel_init=50, # The initial velocity of the split ball
            split_vel_zero_time=0.1, # The time it takes for the speed of the split ball to decay to zero (s)
            stop_zero_time=0.1, # The time to zero the speed after using the stop function
            size_decay_rate=0.0001, # The size proportion of each state frame attenuation
            given_acc_weight=10, # The ratio of actual acceleration to input acceleration
        ))
        return EasyDict(cfg)

    def __init__(self, team_name, name, position, border, size=None, vel=None, acc=None, 
                 vel_last=None, acc_last=None, last_given_acc=None, stop_flag=False, 
                 owner=None, spore_settings=SporeBall.default_config(), **kwargs):
        # init other kwargs
        kwargs = EasyDict(kwargs)
        cfg = CloneBall.default_config()
        cfg = deep_merge_dicts(cfg, kwargs)
        super(CloneBall, self).__init__(name, position, border, size=size, vel=vel, acc=acc, **cfg)
        self.vel_max = cfg.vel_max
        self.acc_max = cfg.acc_max
        self.radius_min = cfg.radius_min
        self.radius_max = cfg.radius_max
        self.radius_init = cfg.radius_init
        self.part_num_max = cfg.part_num_max
        self.on_thorns_part_num = cfg.on_thorns_part_num
        self.on_thorns_part_radius_max = cfg.on_thorns_part_radius_max
        self.split_radius_min = cfg.split_radius_min
        self.eject_radius_min = cfg.eject_radius_min
        self.recombine_age = cfg.recombine_age
        self.split_vel_init = cfg.split_vel_init
        self.split_vel_zero_time = cfg.split_vel_zero_time
        self.stop_zero_time = cfg.stop_zero_time
        self.size_decay_rate = cfg.size_decay_rate
        self.given_acc_weight = cfg.given_acc_weight
        self.spore_settings = spore_settings
        self.cfg = cfg
        # normal kwargs
        self.team_name = team_name
        self.owner = owner
        self.split_acc_init = self.split_vel_init / self.split_vel_zero_time # Initialized deceleration scalar after split
        self.age = 0 # The time of the current ball from the last split

        self.vel_last = Vector2(0, 0) if vel_last is None else vel_last
        self.acc_last = Vector2(0, 0) if acc_last is None else acc_last 

        if self.vel_last.length() > 0:
            self.cooling_last = True
        else:
            self.cooling_last = False
        if not hasattr(self, 'direction'):
            if (self.vel + self.vel_last).length() != 0:
                self.direction = copy.deepcopy((self.vel + self.vel_last).normalize())
            else:
                self.direction = Vector2(0.000001, 0.000001).normalize()

        self.check_border()
        self.stop_flag = stop_flag
        self.last_given_acc = Vector2(0, 0) if last_given_acc is None else last_given_acc

        if not hasattr(self, 'stop_time'):
            self.stop_time = 0
        if not hasattr(self, 'acc_stop'):
            self.acc_stop = Vector2(0, 0)

    def cal_vel_max(self, radius):
        return self.vel_max*20/(radius+10)

    def move(self, given_acc=None, given_acc_center=None, duration=0.05):
        """
        Overview:
            Realize the movement of the ball, pass in the direction and time parameters
        """
        self.age += duration
        if given_acc is not None:
            given_acc = given_acc if given_acc.length() < 1 else given_acc.normalize()
            given_acc *= 10
            self.last_given_acc = given_acc
        else:
            given_acc = self.last_given_acc
        if self.stop_flag: # Stop function used
            if not hasattr(self, 'stop_time'):
                self.stop_time = 0
                self.acc_stop = Vector2(0, 0)
            if self.stop_time < self.stop_zero_time: # Deceleration state
                self.vel = self.vel + self.acc_stop * duration # Acceleration is already in the opposite direction of speed
                self.vel_last += self.acc_last * duration
                self.stop_time += duration
                if self.stop_time >= self.stop_zero_time: # If the stop time is exceeded, set the speed and acceleration to 0 directly
                    self.vel = Vector2(0, 0)
                    self.acc_stop = Vector2(0, 0)
                    self.vel_last = Vector2(0, 0)
                    self.acc_last = Vector2(0, 0)
                self.position = self.position + (self.vel + self.vel_last) * duration
            else: # Exceed the stop time, move closer to the center
                if given_acc_center is None: # Single ball
                    return
                else: # Multiple balls
                    self.acc = format_vector(given_acc*self.acc_max, self.acc_max)
                    acc_tmp = format_vector(self.acc + given_acc_center/math.sqrt(self.radius), self.acc_max) # The acceleration towards the center of mass is handled separately
                    self.vel_max_ball = self.cal_vel_max(self.radius)
                    self.vel = format_vector(self.vel * 0.95 + (self.acc + acc_tmp) * duration, self.vel_max_ball) # vel is multiplied by a number to prevent circling phenomenon
                    self.position = self.position + self.vel * duration
        else: # normal status
            self.acc_stop = Vector2(0, 0)
            if given_acc_center is None:
                given_acc_center = Vector2(0, 0)
            self.acc = format_vector(given_acc*self.acc_max, self.acc_max)
            acc_tmp = format_vector(self.acc + given_acc_center/math.sqrt(self.radius), self.acc_max) # The acceleration towards the center of mass is handled separately
            self.vel_max_ball = self.cal_vel_max(self.radius)
            self.vel = format_vector(self.vel + (self.acc + acc_tmp) * duration, self.vel_max_ball)
            if self.cooling_last:
                self.vel_last += self.acc_last * duration
                if self.age >= self.split_vel_zero_time:
                    self.vel_last = Vector2(0, 0)
                    self.acc_last = Vector2(0, 0)
                    self.cooling_last = False
            self.position = self.position + (self.vel + self.vel_last) * duration

        if (self.vel + self.vel_last).length() != 0:
            self.direction = copy.deepcopy((self.vel + self.vel_last).normalize())
        else:
            self.direction = Vector2(0.000001, 0.000001).normalize()
        self.check_border()

    def eat(self, ball, clone_num=None):
        """
        Parameters:
            clone_num <int>: The total number of balls for the current player
        """
        if isinstance(ball, SporeBall) or isinstance(ball, FoodBall) or isinstance(ball, CloneBall):
            self.set_size(add_size(self.size, ball.size))
            if self.radius > self.radius_max:
                self.radius = self.radius_max
        elif isinstance(ball, ThornsBall):
            assert clone_num is not None
            self.set_size(add_size(self.size, ball.size))
            if clone_num < self.part_num_max:
                split_num = min(self.part_num_max - clone_num, self.on_thorns_part_num)
                return self.on_thorns(split_num=split_num)
        else:
            logging.debug('CloneBall can not eat {}'.format(type(ball)))
        self.check_border()
        return True

    def on_thorns(self, split_num) -> list:
        '''
        Overview:
            Split after encountering thorns, calculate the size, position, speed, acceleration of each ball after splitting
        Parameters:
            split_num <int>: Number of splits added
        Returns:
            Return a list that contains the newly added balls after the split, the distribution of the split balls is a circle and the center of the circle has a ball
        '''
        # middle ball
        around_radius = min(math.sqrt(self.size / (split_num + 1)), self.on_thorns_part_radius_max)
        around_size = around_radius * around_radius
        middle_size = self.size - around_size * split_num
        self.set_size(middle_size) 
        middle_position = Vector2(self.position.x, self.position.y)
        around_positions = []
        around_vels = []
        around_accs = []
        for i in range(split_num):
            angle = 2*math.pi*(i+1)/split_num
            unit_x = math.cos(angle)
            unit_y = math.sin(angle)
            vel = Vector2(self.split_vel_init*unit_x, self.split_vel_init*unit_y)
            acc = - Vector2(self.split_acc_init*unit_x, self.split_acc_init*unit_y)
            around_position = self.position + Vector2((self.radius+around_radius)*unit_x, (self.radius+around_radius)*unit_y)
            around_vels.append(vel)
            around_accs.append(acc)
            around_positions.append(around_position)
        balls = []
        for p, v, a in zip(around_positions, around_vels, around_accs):
            around_ball = CloneBall(team_name=self.team_name, name=uuid.uuid1(), position=p, border=self.border, 
                                    size=around_size, vel=copy.deepcopy(self.vel), acc=copy.deepcopy(self.acc),
                                    vel_last=v, acc_last=a, last_given_acc=copy.deepcopy(self.last_given_acc), 
                                    stop_flag=self.stop_flag, owner=self.owner, spore_settings=self.spore_settings, 
                                    **self.cfg)
            balls.append(around_ball)
        return balls

    def eject(self, direction=None) -> list:
        '''
        Overview:
            When spit out spores, the spores spit out must be in the moving direction of the ball, and the position is tangent to the original ball after spitting out
        Returns:
            Return a list containing the spores spit out
        '''
        if direction is None:
            direction = copy.deepcopy(self.direction)
        if self.radius >= self.eject_radius_min:
            spore_radius = self.spore_settings.radius_min
            self.set_size(self.size - spore_radius**2)
            direction_unit = direction.normalize()
            position = self.position + direction_unit * (self.radius + spore_radius)
            return SporeBall(name=uuid.uuid1(), position=position, border=self.border, direction=direction_unit, **self.spore_settings)
        else:
            return False

    def split(self, clone_num, direction=None) -> list:
        '''
        Overview:
            Active splitting, the two balls produced by splitting have the same volume, and their positions are tangent to the forward direction
        Parameters:
            clone_num <int>: The total number of balls for the current player
        Returns:
            The return value is the new ball after the split
        '''
        if direction is None:
            direction = copy.deepcopy(self.direction)
        if self.radius >= self.split_radius_min and clone_num < self.part_num_max:
            split_size = self.size / 2
            self.set_size(split_size)
            clone_num += 1
            direction_unit = direction.normalize()
            position = self.position + direction_unit * (self.radius * 2)
            vel_split = self.split_vel_init * direction_unit
            acc_split = - self.split_acc_init * direction_unit
            return CloneBall(team_name=self.team_name, name=uuid.uuid1(), position=position, border=self.border, 
                             size=split_size, vel=copy.deepcopy(self.vel), acc=copy.deepcopy(self.acc),
                             vel_last=vel_split, acc_last=acc_split, last_given_acc=copy.deepcopy(self.last_given_acc),
                             stop_flag=self.stop_flag, owner=self.owner, spore_settings=self.spore_settings, **self.cfg)
        else:
            return False

    def rigid_collision(self, ball):
        '''
        Overview:
            When two balls collide, We need to determine whether the two balls belong to the same player
            A. If not, do nothing until one party is eaten at the end
            B. If the two balls are the same owner, judge whether the age of the two is full or not meet the fusion condition, if they are satisfied, do nothing.
            C. If the two balls are the same owner, judge whether the age of the two is full or not meet the fusion condition, Then the two balls will collide with rigid bodies
            This function completes the C part: the rigid body collision part, the logic is as follows:
             1. To determine the degree of fusion of the two balls, use [the radius of both] and subtract [the distance between the two] as the magnitude of the force
             2. Calculate the coefficient according to the weight, the larger the weight, the smaller the coefficient will be
             3. Correct the position of the two according to the coefficient and force
        Parameters:
            ball <CloneBall>: another ball
        Returns:
            state <bool>: the operation is successful or not
        '''
        if ball.name == self.name:
            return True
        assert isinstance(ball, CloneBall), 'ball is not CloneBall but {}'.format(type(ball))
        assert self.owner == ball.owner
        assert self.age < self.recombine_age or ball.age < ball.recombine_age
        p = ball.position - self.position
        d = p.length()
        if self.radius + ball.radius > d:
            f = min(self.radius + ball.radius - d, (self.radius + ball.radius - d) / (d+0.00001))
            self.position = self.position - f * p * (ball.size / (self.size + ball.size))
            ball.position = ball.position + f * p * (self.size / (self.size + ball.size))
        else:
            print('WARNINGS: self.radius ({}) + ball.radius ({}) <= d ({})'.format(self.radius, ball.radius, d))
        self.check_border()
        ball.check_border()
        return True

    def stop(self, direction=None):
        '''
        Overview:
            Stop in place, the speed and acceleration become 0, and the direction points to the center of gravity of the shape surrounded by the centers of all spheres
            Acceleration is set to the opposite direction of speed
            stop_zero_time = 0.2
            stop_flag = True
            Judge stop_flag at the next move
            1. True: The direction must be the center of the circle, the acceleration and the upper limit of the speed are reduced
            2. False: Normal movement
        Parameters:
            direction <Vector2>: The direction of the speed after the stop
        Returns:
            state <bool>:  the operation is successful or not
        '''
        self.stop_flag = True
        self.stop_time = 0
        self.acc_stop = - 1 / self.stop_zero_time * (self.vel + self.vel_last)
        self.acc = Vector2(0, 0)
        self.direction = direction if direction is not None else Vector2(0.000001, 0.000001)
        self.last_given_acc = Vector2(0, 0)
        return True

    def judge_rigid(self, ball):
        '''
        Overview:
            Determine whether two balls will collide with a rigid body
        Parameters:
            ball <CloneBall>: another ball
        Returns:
            <bool>: collide or not
        '''
        return self.age < self.recombine_age or ball.age < ball.recombine_age

    def check_border(self):
        """
        Overview:
            Check if the position of the ball is beyond the bounds of the map
        """
        if self.position.x < self.border.minx + self.radius or self.position.x > self.border.maxx - self.radius:
            self.position.x = max(self.position.x, self.border.minx + self.radius)
            self.position.x = min(self.position.x, self.border.maxx - self.radius)
            self.vel.x = 0
            self.acc.x = 0
            self.acc_last.x = 0 

        if self.position.y < self.border.miny + self.radius or self.position.y > self.border.maxy - self.radius:
            self.position.y = max(self.position.y, self.border.miny + self.radius)
            self.position.y = min(self.position.y, self.border.maxy - self.radius)
            self.vel.y = 0
            self.acc.y = 0
            self.acc_last.y = 0 

    def size_decay(self):
        '''
        Overview: 
            Control the size of the ball to decay over time
        '''
        self.set_size(self.size * (1-self.size_decay_rate))
        return True


    def __repr__(self) -> str:
        return '{}, vel_last={}, acc_last={}, age={:.3f}, owner={}, direction={}, team_name={}'\
                .format(super().__repr__(), self.vel_last, self.acc_last, self.age, self.owner, self.direction, self.team_name)



import math
import random
import numpy as np


def chunks(arr, m):
    n = int(math.ceil(len(arr) / float(m)))
    return [arr[i:i + n] for i in range(0, len(arr), n)]

def get_probability(src,arr):
    diff = [abs(i-src)+0.001 for i in arr]
    return [1/i if 1/i<1 else 1 for i in diff]

def norm(arr):
    return [i/sum(arr) for i in arr]

def to_aliased_circle(position, radius, cut_num=16, decrease=3):
    point_list = []
    radius_decrease = radius - decrease
    assert radius_decrease > 0
    piece_angle = math.pi / (cut_num)
    for i in range(cut_num*2):
        angle = piece_angle * i
        if i % 2 == 0:
            point_list.append([position.x + radius * math.cos(angle), position.y + radius * math.sin(angle)])
        else:
            point_list.append([position.x + radius_decrease * math.cos(angle), position.y + radius_decrease * math.sin(angle)])
    return point_list



import os
import pygame
import platform


class BaseRender:

    def __init__(self, width, height, padding=(0,0), cell_size=10, only_render=False):
        pygame.init()
        if platform.system() == 'Linux': # If the current system is linux, window is not used
            os.environ["SDL_VIDEODRIVER"] = "dummy"
        self.width = width
        self.height = height
        self.padding = padding
        self.width_full = self.width + self.padding[0] * 2
        self.height_full = self.height + self.padding[1] * 2
        self.cell_size = cell_size
        self.FPS = 60 # Set the frame rate (the number of screen refreshes per second)
        self.fpsClock = pygame.time.Clock() 
        if not only_render:
            self.screen = pygame.display.set_mode((self.width_full, self.height_full),  0, 32)
            pygame.display.set_caption("GoBigger - OpenDILab Environment")

    def fill(self, server):
        raise NotImplementedError

    def show(self):
        raise NotImplementedError

    def close(self):
        raise NotImplementedError
